/* 
  File: anneal.c

   This is a working version of 'flexible' code.  File output, timing, 
   initial beta, initial input (, etc. ) controlled dynamically from 
   command line.

   Annealing schedules:
   
     EXP:       T(k+1) = alpha * T(K)

     ADAPTIVE:  T(k+1) = T(k) * exp [ -lambda * T(k) / sd ( E ) ],

                where sd ( E ) is the standard deviation of the ensemble
		of energies of size Ensemble.

		
   Measures overlap of final configurations of consequtive betas.

 */

#include "main.h"

int    Cur_Beta_Index = 0;                      /* */

FILE   *Efile;
 
extern int        
       Num_Reject[MAX_BETAS],
       Num_Accept[MAX_BETAS],

       Rows, Cols,
       NTherm, Num_Sweeps, Freq,
       Measurements;


main(argc, argv)
int  argc;
char  *argv[];
{

   int cur_sub_beta,
       initial_conf,     
       n,
       model,
       seed,
       temp_spins[MAXROWS][MAXCOLS];

   clock_t min, sec;  /* clock_t is an int */

   double beta,
          new_beta,
          g_diff,
          beta_diff,
          min_beta,
          overlap_coeff,

          sub_beta[MAX_SUB_BETAS],
          avg_energy[MAX_SUB_BETAS];

   FILE *DGfile, *SDfile, *Afile, *Rfile, *beta_file, *overlap_file;

   Get_Switches( argc, argv );

   Input_Parameters( &initial_conf, &n,  &seed, &model );

/* Initialize random numbers */
   srand48( seed );

   Min_Energy[Cur_Beta_Index] = 1000.0;

   Initialize ( &beta, initial_conf, model );

   if ( DUMP_DG )
   {
     if ( (DGfile = fopen( "dat.dg", "w" )) == NULL ) Exit();
     
     fprintf( DGfile, "%d\n%d\n", Rows, Cols );
     fprintf( DGfile, "%d\n%d\n%d\n", NTherm, Num_Sweeps, Freq );
     fflush ( DGfile );
   }

   if ( DUMP_SD )
     if ( (SDfile = fopen( "dat.sd", "w" )) == NULL ) Exit();

   if ( DUMP_BETAS )
     if ( (beta_file = fopen( "dat.beta", "w" )) == NULL ) Exit();

   if ( DUMP_ACCEPT )
     if ( (Afile = fopen( "dat.accept", "w" )) == NULL ) Exit();

   if ( DUMP_REJECT )
     if ( (Rfile = fopen( "dat.reject", "w" )) == NULL ) Exit();

   if ( DUMP_ENERGY )
     if ( (Efile = fopen( "dat.energy", "w" )) == NULL ) Exit();

   if ( DUMP_OVERLAP )
     if ( (overlap_file = fopen( "dat.overlap", "w" )) == NULL ) Exit();

   overlap_coeff = 0.0;
   One_Pass( beta, &min_beta, overlap_coeff );

   printf( "\nInitital info ->\n" );
   printf( "\t\tBeta_0 = %lf\n", beta );
   printf( "\t\tEnergy = %lf \t Accept = %lf \n", Total_Energy, 
	   Total_Accept/Num_Spins * 100.0 );

   printf( "\n\nBeta No.: Beta: tot(energy): err(energy) ->\n" );

   if ( DUMP_ACCEPT ) 
   {
     fprintf( Afile, "%lf %lf \n", beta, (Total_Accept / Num_Spins * 100.0) );
     fflush ( Afile );
   }

/* start clock */
   if ( TIMING ) clock();

   for ( Cur_Beta_Index = 0; Cur_Beta_Index < MAX_BETAS; Cur_Beta_Index ++ ) {

     sub_beta[0] = beta;
     avg_energy[0] = Total_Energy;
     g_diff = Total_Energy;

     if ( DUMP_OVERLAP )
       Save_Spins( temp_spins );

     new_beta = Get_New_Beta( beta );

     beta_diff = new_beta - beta;
     beta_diff /= (double)n;

     Num_Reject[Cur_Beta_Index] = 0;
     Num_Accept[Cur_Beta_Index] = 0;

     for ( cur_sub_beta = 1; cur_sub_beta <= n; cur_sub_beta++ ) {

       sub_beta[cur_sub_beta] = beta + beta_diff * cur_sub_beta;

       One_Pass( sub_beta[cur_sub_beta], &min_beta, overlap_coeff );

       printf("%d: %16.8f \t %16.8f \t %16.8f\n", Cur_Beta_Index,
	      sub_beta[cur_sub_beta], Total_Energy, Error_Energy );
       fflush(stdout);

       if ( DUMP_ACCEPT ) {
	 fprintf( Afile, "%lf %lf \n", sub_beta[cur_sub_beta], 
		  (Total_Accept / Num_Spins * 100.0) );
	 fflush ( Afile );
       }

       avg_energy[cur_sub_beta] = Total_Energy;
       g_diff += Total_Energy;
     } /* end of loop over beta SUBDIVISIONS */

     g_diff = (g_diff - 0.5*(avg_energy[0]+avg_energy[n])) * beta_diff; 
     
     printf( "E_(n) = %lf E_(n+1) = %lf Delta(g) = %lf\n",
	    avg_energy[0], avg_energy[n], g_diff );

     if ( DUMP_REJECT ) {
       fprintf( Rfile, "%lf %d %d\n", sub_beta[cur_sub_beta], 
	        Num_Reject[Cur_Beta_Index], Num_Accept[Cur_Beta_Index] );
       fflush ( Rfile );
     }


     if ( DUMP_OVERLAP ) {
       overlap_coeff = Calc_Overlap( temp_spins, Spins );
       fprintf( overlap_file, "%lf %lf\n", beta, overlap_coeff );
       fflush ( overlap_file );
       printf( "%lf\n", overlap_coeff );
     }
       
     if ( DUMP_DG ) {
       fprintf( DGfile, "%.8f\n", beta );
       fprintf( DGfile, "%.8f\n", g_diff );
       fflush ( DGfile );
     }
  
     if ( DUMP_SD ) {
       fprintf( SDfile, "%lf %lf\n", beta, Std_Dev );
       fflush ( SDfile );
     }

     if ( DUMP_BETAS ) {
       fprintf( beta_file, "%d %lf\n", Cur_Beta_Index, beta );
       fflush ( beta_file );
     }

     if ( DUMP_OVERLAP ) {
       fprintf( overlap_file, "%lf %lf\n", beta, overlap_coeff );
       fflush ( overlap_file );
     }
	       
     beta = new_beta;

     if ( Cur_Beta_Index + 1 < MAX_BETAS ) 
       Min_Energy[Cur_Beta_Index + 1] = Min_Energy[Cur_Beta_Index];

     if ( Terminate( Step ) ) break;

   } /* Loop over betas */

   /* stop clock */
   if ( TIMING ) {
     sec = clock() / CLOCKS_PER_SEC;
     min = sec / 60;
     printf( "\nElapsed time (min:sec)... %d:%d\n\n", min, (sec % 60) );
   }

   if ( Cur_Beta_Index == MAX_BETAS ) Cur_Beta_Index --;

   printf( "\nFinal info -> \n" );
   printf( "\t\tMinimum Energy = %16.8lf at Beta = %lf\n", 
	   Min_Energy[Cur_Beta_Index], min_beta ); 
   printf( "\t\tNumber of betas visited = %d\n", Cur_Beta_Index + 1 );
   printf( "\t\tTotal iterations = %d\n", Get_Iter( n ) );

   if ( DUMP_DG )
   {
     fprintf ( DGfile, "%.8f\n0.0\n-1\n", beta );
     fprintf ( DGfile, "%d\n%d\n%d\n", initial_conf, model, seed );
     fflush  ( DGfile );
   }

   if ( DUMP_SD ) {
     fprintf( SDfile, "%lf %lf\n", beta, Std_Dev );
     fflush ( SDfile );
   }

}   /*  main ()  */
          


void Exit()
{
  printf( "\nError opening file.  Exiting...\n" );
  exit(2);
}
