/*
  File: init.c

 */

#include <string.h>
#include "global.h"
#include "switches.h"


int    Rows, Cols, Num_Spins,
       Freq,
       Num_Sweeps,
       NTherm;

void   Input_Parameters ();

void Input_Parameters( Initial_Conf, n, Seed, Model )
int *Initial_Conf;
int *n, *Seed, *Model;
{
   int i; 
   char Sentinel;

   do   
   {
      Sentinel = 0;
      printf ( "Enter the vertical dimensions of the lattice (0 < Y <= %d): ",
	       MAXROWS );
      fscanf ( Init_File, "%d", &Rows );

      if ( ( Rows > MAXROWS ) || ( Rows <= 0 ) )
      { 
         printf ( "\n\nInput out of range.\n\n" );
         Sentinel = 1;
      }
   } 
   while (Sentinel);
   printf("%d", Rows);

   do
   {
      Sentinel = 0;

      printf("\n\n");
      printf ( "Enter the horizontal dimensions of the lattice (0 < X <= %d): ",
	       MAXCOLS );
      fscanf ( Init_File, "%d", &Cols );

      if ( ( Cols > MAXCOLS ) || ( Cols <= 0 ) )
      {
            printf ( "\n\nInput out of range.\n\n" );
            Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", Cols);

   Num_Spins = Rows * Cols;

   printf ( "\n\nLATTICE SIZE :    %d X %d = %d\n", Rows, Cols, Num_Spins );

   printf ( "\n\nEnter the number of thermalization sweeps desired: " );
   fscanf ( Init_File, "%d", &NTherm );
   if ( NTherm < 0 ) NTherm = 0;
   printf("%d", NTherm);

   do
   {
      Sentinel = 'n';        

      printf ( "\nEnter the total number of calculation updates \n" );
      printf ( "desired for each Beta: : " );
      fscanf ( Init_File, "%d", &Num_Sweeps );
      if ( Num_Sweeps <= 0 )   
         {
            printf ( "\n\nInvalid input\n\n" );
            Sentinel = 'y';
         }
   } while (Sentinel == 'y');
   printf("%d", Num_Sweeps);

   do
   { 
      Sentinel = 'n';

      printf ( "\nEnter the number of sweeps per calculation update: ");
      fscanf ( Init_File, "%d", &Freq );
      if ( Freq > Num_Sweeps ) Freq = Num_Sweeps;
      if ( Freq <= 0 ) 
      { 
         printf ( "\n\nInvalid input\n\n" );
         Sentinel = 'y';
      }   
   } while ( Sentinel == 'y' ); 
   printf("%d", Freq);

   do
   {  
      printf ( "\nEnter the Grain Size for Calculating Delta_G: " );
      fscanf ( Init_File, "%d", n );
   }
   while ( *n <= 0 );
   printf("%d", *n);

   do 
   {
      Sentinel = 0;
      printf ( "\n\nEnter the number next to your choice of spin\n" );
      printf ( "arrangements for the initial lattice.\n\n" );
      printf ( "1 -- All spins up.\n" );
      printf ( "2 -- All spins down.\n" );
      printf ( "3 -- Random configuration.\n\n\n" );
      printf ( "Choice: " );
      fscanf ( Init_File, "%d", Initial_Conf );

      if ((*Initial_Conf != 1) && (*Initial_Conf != 2) && (*Initial_Conf != 3))
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Initial_Conf);
   
   do 
   {
      Sentinel = 0;
      printf ( "\n\nWhich model do you want to simulate?\n\n" );
      printf ( "1 -- Ferromagnet.\n" );
      printf ( "2 -- Spin Glass.\n" );
      printf ( "Choice: " );
      fscanf ( Init_File, "%d", Model );

      if ( (*Model != 1) && (*Model != 2) )
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Model);

   printf ( "\n\nInput a seed for the random number generator: " );
   fscanf ( Init_File, "%d", Seed );
   printf("%d\n\n", *Seed);
   fflush(stdout);

}


double t_rand ();
void   Init_Lattice ();

int    Spins[MAXROWS][MAXCOLS],    /* spins */
       J[MAXROWS][MAXCOLS][2];     /* couplings */


void  Init_Lattice( Initial_Conf, Model )
int Initial_Conf, Model;
{
   int i, j, dim;

   switch (Initial_Conf)
   {

      case 1  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = 1;
           break;

      case 2  : 
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = -1;
           break;

      case 3  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 if ( t_rand() < 0.5  )
                    Spins[i][j] = 1;
                 else 
                    Spins[i][j] = -1;
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }

   switch (Model)
   {

      case 1  :  /* ferromagnet */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    J[i][j][dim] = 1;
           break;

      case 2  :  /* spin glass */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    if ( t_rand() < 0.5  )
                       J[i][j][dim] = 1;
                    else 
                       J[i][j][dim] = -1;
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }
}


double Init_Beta ();

void Initialize ( Beta, initial_conf, model )
/*
  Output: If AUTO_START is turned on, this routine returns a Beta value
          whose acceptance is approximately 50%.
 */
double *Beta;
int    initial_conf,
       model;
{

  Init_Lattice ( initial_conf, model );
  *Beta = New_Beta;

  if ( ! AUTO_START )
    return;

  *Beta = Init_Beta ( *Beta, initial_conf, model );

} /* Initialize ( ) */


