/*
  File: switches.c

 */

#include <string.h>
#include <stdlib.h>
#include "switches.h"


int    Step     = 3;
int    Ensemble = 50;
double Alpha    = 0.8;
double Lambda   = 10.0;
double New_Beta = BETA;

int AUTO_START  = OFF;
int TIMING      = OFF;
int CONV        = OFF;
int DUMP_DG     = OFF;
int DUMP_SD     = OFF;
int DUMP_ACCEPT = OFF;
int DUMP_REJECT = OFF;
int DUMP_ENERGY = OFF;
int DUMP_OVERLAP= OFF;
int DUMP_BETAS  = OFF;
int SCHED       = EXP;

FILE *Init_File = stdin;

void Get_Switches ( argc, argv )
/*
  Extracts command line switches for customized execution.  No command
  line arguments causes a summary of possible switches to be displayed.
 */
int  argc;
char **argv;
{

  int i = 0;

  if ( argc == 1 ) 
    defaults();

  printf( "\n\nRuntime parameters -> \n" );

  while ( ++ i < argc )
  {

    if ( strcmp( argv[i], "-all" ) == 0 )
    {
      printf( "\nGenerating all data files...\n" );
      printf( "\tAcceptance at each beta\n" );
      printf( "\tEnergy blah blah blah\n" );
      DUMP_DG     = ON;
      DUMP_SD     = ON;
      DUMP_ACCEPT = ON;
      DUMP_REJECT = ON;
      DUMP_ENERGY = ON;
      DUMP_BETAS  = ON;
      TIMING      = ON;
    }

    else if ( strcmp( argv[i], "-auto" ) == 0 )
    {
      printf( "\tDynamic Beta_0 determination \n" );
      AUTO_START = ON;
    }

    else if ( strcmp( argv[i], "-dg" ) == 0 )
    {
      printf( "\tGenerating Delta(G) data file \n" );
      DUMP_DG = ON;
    }

    else if ( strcmp( argv[i], "-b" ) == 0 )
    {
      printf( "\tGenerating Beta data file \n" );
      DUMP_BETAS = ON;
    }

    else if ( strcmp( argv[i], "-sd" ) == 0 )
    {
      printf( "\tGenerating Standard Deviation data file \n" );
      DUMP_SD = ON;
    }

    else if ( strcmp( argv[i], "-accept" ) == 0 )
    {
      printf( "\tGenerating Acceptance data file\n" );
      DUMP_ACCEPT = ON;
    }

    else if ( strcmp( argv[i], "-overlap" ) == 0 )
    {
      printf( "\tGenerating Overlap data file \n" );
      DUMP_OVERLAP = ON;
    }

    else if ( strcmp( argv[i], "-reject" ) == 0 )
    {
      printf( "\tGenerating Rejection data file \n" );
      DUMP_REJECT = ON;
    }

    else if ( strcmp( argv[i], "-energy" ) == 0 )
    {
      printf( "\tGenerating Energy data file \n" );
      DUMP_ENERGY = ON;
    }

    else if ( strcmp( argv[i], "-adapt" ) == 0 )
    {
      printf( "\tUsing Adaptive annealing schedule\n" );
      SCHED = ADAPT;
    }

    else if ( strcmp( argv[i], "-ensemble" ) == 0 )
    {
      Ensemble = atoi( argv[++i] );
      printf( "\tUsing new number of Configurations for sd calculation " );
      printf( "(%d)\n", Ensemble );
    }

    else if ( strcmp( argv[i], "-time" ) == 0 )
    {
      TIMING = ON;
      printf("\tTiming enabled \n" );
    }

    else if ( strcmp( argv[i], "-conv" ) == 0 )
    {
      CONV = ON;
      printf("\tDisplay of convergence criteria enabled \n" );
    }

    else if ( strcmp( argv[i], "-step" ) == 0 )
    {
      Step = atoi( argv[++i] );
      printf( "\tUsing new Step size (%d) \n", Step );
    }

    else if ( strcmp( argv[i], "-beta" ) == 0 )
    {
      New_Beta = atof( argv[++i] );
      printf( "\tUsing new Beta (%lf) \n", New_Beta );
    }

    else if ( strcmp( argv[i], "-alpha" ) == 0 )
    {
      Alpha = atof( argv[++i] );
      printf( "\tUsing new Alpha (%lf) \n", Alpha );
    }

    else if ( strcmp( argv[i], "-lambda" ) == 0 )
    {
      Lambda = atof( argv[++i] );
      printf( "\tUsing new Lambda (%lf) \n", Lambda );
    }

    else if ( strcmp( argv[i], "-infile" ) == 0 )
    {
      if ( (Init_File = fopen( argv[++i], "r" )) == NULL )
      {
	printf( "Error opening file.  Exiting...\n" );
	exit(0);
      }
      printf( "\tUsing initial parameters from file %s \n", argv[i] );
    }

    else 
      defaults();

  }
  
  printf( "\n\n" );

}  /* Get_Switches () */


void defaults ()
/*
 */
{

  printf( "\nUsage: \n" );
  printf( "\t anneal <[options]> \n\n" );
  printf( "\t -all      \t Enable all output and timing\n" );
  printf( "\t -infile <input file> \n" );
  printf( "     \t\t\t Use input parameters form file <input file> \n" );
  printf( "\t -auto     \t Dynamic beta_0 determination... \n" );
  printf( "\t -dg       \t Generate Delta(G) data file\n" );
  printf( "\t -b        \t Generate Beta data file\n" );
  printf( "\t -sd       \t Generate Standard Deviation data file\n" );
  printf( "\t -accept   \t Generate Acceptance data file\n" );
  printf( "\t -overlap  \t Generate Overlap data file\n" );
  printf( "\t -reject   \t Generate Rejection data file\n" );
  printf( "\t -energy   \t Generate Energy data file\n" );
  printf( "\t -adapt    \t Use Adaptive annealing schedule\n" );
  printf( "\t -ensemble <num steps> \n" );
  printf( "     \t\t\t Number of Configurations for sd calculation\n" );
  printf( "\t -beta   <beta not>  \n \t\t\t Change beta not\n" );
  printf( "\t -alpha  <exp fact> \n" );
  printf( "     \t\t\t Change default exp schedule factor\n" );
  printf( "\t -lambda <adapt fact> \n" );
  printf( "     \t\t\t Change default adaptive schedule factor\n" );
  printf( "\t -step   <step size> \n \t\t\t Change default Step size\n" );
  printf( "\t -time     \t Enable Timing\n" );
  printf( "\t -conv     \t Enable display of Convergence information\n" );
  
  exit (0);
} /* defaults ( ) */



