! A.13 2-Dimensional Convolution

! The convolution of two images `a' and `b' are stored in the
! matrix `c', and is calculated by the steps: 

!  1.Perform 2-D FFT for Image `a'.
!  2.Perform 2-D FFT for Image `b'.
!  3.Perform matrix multiplication c = FFT(a)  FFT(b).
!  4.Perform inverse 2-D FFT on `c' to obtain the convolution 

! where each 2-D FFT is a combination of parallel 1-D FFTs
! which are solved in the same manner as codeA.11.2 above. 

      PROGRAM convolve
      IMPLICIT NONE
      INCLUDE 'convolve.inc'
C Define interface for subroutine fftrow.
       INTERFACE
        PURE SUBROUTINE fft_slice(a,N,sign)
          INTEGER,INTENT(IN)                 :: N,sign
          COMPLEX,DIMENSION(N),INTENT(INOUT) :: a
        END SUBROUTINE fft_slice
      END INTERFACE
C Define interface for subroutine timer.
      INTERFACE
        SUBROUTINE timer(return_time, initial_time)
          REAL, INTENT(IN)  :: initial_time
          REAL, INTENT(OUT) :: return_time
        END SUBROUTINE timer
      END INTERFACE
      COMPLEX, DIMENSION(NN,NN) :: a,b,c
      INTEGER                   :: icol,sign
      REAL                      :: norm
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms, tcomp
C Distribute arrays.

! These HPF statements start out like all the others, by declaring
! the 1D abstract processor array procs, of size NUMBER_OF_PROCESSORS(),
! which is determined at run time.  A 2D template is declared ,
! and it is distributed by Blocks along its second dimension.  Its
! first dimension is collapsed for the distribution.  The other
! arrays are aligned with the template, and are thus distributed in the
! same fashion.

!HPF$ PROCESSORS procs(NUMBER_OF_PROCESSORS())
!HPF$ TEMPLATE        :: temp(NN,NN)
!HPF$ DISTRIBUTE(*,BLOCK) ONTO procs :: temp
!HPF$ ALIGN WITH temp :: a,b,c
      INTRINSIC TRANSPOSE
      INTRINSIC REAL,AIMAG
C Begin Executable
C Start timer
      CALL timer(ti, 0.0)
      tcomp = 0.0
      tcomms = 0.0
      norm = 1.0/(NN*NN)
C Initialise the arrays.
      a = (1.0,0.0)
      b = (0.0,1.0)
C Set parameter to tell fft_slice to do FFT transform.
      sign = 1
      CALL timer(tstart, 0.0)
!HPF$ INDEPENDENT
      DO icol = 1,NN
        CALL fft_slice(a(:,icol),NN,sign)
        CALL fft_slice(b(:,icol),NN,sign)
      END DO
      CALL timer(tend, tstart)
      tcomp = tcomp + tend
C.. Transpose
      CALL timer(tstart, 0.0)
      a = TRANSPOSE(a)
      b = TRANSPOSE(b)
      CALL timer(tend, tstart)
      tcomms = tcomms + tend
      CALL timer(tstart, 0.0)
!HPF$ INDEPENDENT
      DO icol = 1,NN
        CALL fft_slice(a(:,icol),NN,sign)
        CALL fft_slice(b(:,icol),NN,sign)
      END DO
      c = a*b
C Calculate the inverse of `c'
      sign = -1
!HPF$ INDEPENDENT
      DO icol = 1,NN
        CALL fft_slice(c(:,icol),NN,sign)
      END DO
      CALL timer(tend, tstart)
      tcomp = tcomp + tend
C.. Transpose
      CALL timer(tstart, 0.0)
      c = TRANSPOSE(c)
      CALL timer(tend, tstart)
      tcomms = tcomms + tend
      CALL timer(tstart, 0.0)
!HPF$ INDEPENDENT
      DO icol = 1,NN
        CALL fft_slice(c(:,icol),NN,sign)
      END DO
C Normalise.
      c = c * norm
      CALL timer(tend, tstart)
      tcomp = tcomp + tend
C Stop timer
      CALL timer(tf, ti)
C Simple check the FFT has worked.
      WRITE(UNIT=STMOUT,FMT=1001) NN*NN
      WRITE(UNIT=STMOUT,FMT=1002) REAL(c(1,1)),AIMAG(c(1,1))
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),NN,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
      STOP
 1001 FORMAT('c(1,1) should be:   ',I9)
 1002 FORMAT('c(1,1) is actually: ',E15.6,E15.6)
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I5/
     2       'Communications = ',F9.3/
     3       'Compute        = ',F9.3/
     4       'Others         = ',F9.3/
     5       'Total time     = ',F9.3)
      END
C======================================================================



       subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end
