!  A.7 2-D Binary Phase Quenching of
!  Cahn Hilliard Cook Equation

! This code sets up a random binary phase mixture and applies
! the Cahn Hilliard Cook equation to quench the system into a
! cooled state. 

      PROGRAM chc
      IMPLICIT NONE
C The mesh size
      integer, parameter :: n1 = 256
      integer, parameter :: n2 = 256
C number of bins to measure
      integer, parameter :: nb = 20
      real, dimension(1:nb) :: bins
      real, dimension(n1,n2) :: c, f
      real ti, tf,start,tcalc,tcom,a,others
      real cmax,less
! An abstract processor array procs is declared, whose size NUMBER_OF_PROCESSORS()
! is declared at run time.  A 1D template is defined and distrubed by
! blocks onto the processor array. The 2D arrays then have their second dimension
! aligned with the template, insuring that the distribution is the same as 
! the template's.  The first dimension of these arrays is collapsed as 
! far as the distribution goes with the *.

!HPF$ PROCESSORS :: PROCS(NUMBER_OF_PROCESSORS())
!HPF$ TEMPLATE, DIMENSION(n2) :: TEMP
!HPF$ DISTRIBUTE (BLOCK) ONTO PROCS :: TEMP
!HPF$ ALIGN (*,:) WITH TEMP(:) :: c,f
      interface
         subroutine timer(return_time, initial_time)
         real, intent(in)  :: initial_time
         real, intent(out) :: return_time
         end subroutine
      end interface
      real t1, t2, x
      real :: dt, dtby2
      integer :: iter
      integer :: niter
      integer i,j
C Executable Code.
      c = 0.0
      f = 0.0
      bins = 0.0
      OPEN(UNIT=3,FILE='in.runme',STATUS='OLD')
      READ(UNIT=3,FMT=*) dt
      READ(UNIT=3,FMT=*) niter
      dtby2 = dt / 2.0
      WRITE(UNIT=6,FMT='(''DT = '',F9.3,'' niter = '',I12)') dt,niter
      tcalc=0.0
      tcom=0.0
      call timer(start,0.0)
      call timer(ti,0.0)
C Use the (poor) system supplied RNG
      c = 1.0
      call random_number( c )
      c = c - 0.5
      call timer(tcalc,ti)
      do iter=1,niter
        call timer(ti,0.0)
        f = -1.0*(cshift(c,1,1) + cshift(c,-1,1))
        f = 3.0 * c +  c**3 + f
        call timer(a,ti)
        tcalc=tcalc+a
        call timer(ti,0.0)
        f= f- (cshift(c,1,2) + cshift(c,-1,2))
        call timer(a,ti)
        tcom=tcom+a
        call timer(ti,0.0)
        c = c + dtby2 * ( cshift(f,1,1) + cshift(f,-1,1)-4.0*f)
        call timer(a,ti)
        tcalc=tcalc+a
        call timer(ti,0.0)
        c = c + dtby2*(cshift(f,1,2) + cshift(f,-1,2))
        call timer(a,ti)
        tcom=tcom+a
        call timer(ti,0.0)
C be cautious in case of leaks
        cmax = maxval( abs(c) )
        if( cmax > 0.0 ) c = c / cmax
        call timer(a,ti)
        tcalc=tcalc+a
      end do
      call timer(tf,start)
      others=tf-tcalc-tcom
      write(6,6001) NUMBER_OF_PROCESSORS(),
     1              n1,tcom,tcalc,others,tf
 6001 FORMAT('Number of Processors =',I4/
     1 'Problem size =',I4/
     2 'Communications = ',F9.3/
     3 'Compute = ', F9.3/
     4 'Others =', F9.3/
     5 'Total time =',F9.3 )
      STOP
      end program
C======================================================================
      subroutine timer(return_time, initial_time)
      real,  intent(in)  :: initial_time
      real,  intent(out) :: return_time
      integer finish,rate
      CALL system_clock( COUNT=finish,COUNT_RATE=rate)
      return_time= FLOAT(finish) / FLOAT(rate) - initial_time
      RETURN
      end subroutine
C=============================
