! A.3 Binomial Tree Stock Market
! Simulation

! This is a binomial approximation stock market simulation
! model, incorporating stochastic volatility with American call
! (exercise at any time in contract) and with European call
! (exercise only at option maturity). The timing information output lists time spent
! executing statements with communication required, without communication, and
! others.  The no-communciation times should scale nicely, but those with
! communication obviously will not.


      PROGRAM main
      IMPLICIT NONE
C Define interfaces.
      INTERFACE
         SUBROUTINE timer(return_time, initial_time)
         REAL, INTENT(IN)  :: initial_time
         REAL, INTENT(OUT) :: return_time
         END SUBROUTINE timer
         SUBROUTINE asvcorp2(sp,sk,r,t,v,psi,xmu,cor,reps1,reps,div,
     1                       cval,tcomp,tcomms)
         REAL :: sp,sk,r,t,v,psi,xmu,cor,div,cval
         REAL :: tcomp, tcomms
         INTEGER reps,reps1
         END SUBROUTINE asvcorp2
         SUBROUTINE svcorp2(sp,sk,r,t,v,psi,xmu,cor,reps,cval,
     1                      tcomp,tcomms)
         REAL :: sp,sk,r,t,v,psi,xmu,cor,cval
         REAL :: tcomp, tcomms
         INTEGER reps
         END SUBROUTINE svcorp2
      END INTERFACE
      real sz,k,r,t,tdiv,div,v(4),rate
      real psi,xmu,rho,cbscall,camc,ceur,cams
      integer repb,reps,i, j, times, reps1
      REAL :: timediff,rtimes
      REAL :: ti,tf
      REAL :: tstart,tend
      REAL :: tcomms,tcomp
C Begin Executable
C Start timer
      CALL timer(ti, 0.0)
      tcomp  = 0.0
      tcomms = 0.0
      k = 100.00
      sz = 110
      repb = 100
      reps = 17
      psi = 1.0
      xmu = 0.0
      rate = 0.15
      rho = 0.234
      k = 100.00
      r = 4.50
      t = .25
      v(1) = 0.15
      v(2) = 0.15
      v(3) = 0.15
      v(4) = 0.15
      div = 1.0
      tdiv = 0.125
      times = 1
      rtimes = float(times)
c --- stochastic volatility European Model
      print *,'stochastic volatility European Model'
      print *,'--------------------------------------'
      print *,'reps           Value       Time(sec.)'
      do 1111 reps = 17, 17
         CALL timer(tstart, 0.0)
         do 6000 i = 1, times
            call svcorp2(sz,k,rate,t,v(3),psi,xmu,rho,reps,ceur,
     1                   tcomp,tcomms)
 6000    continue
         CALL timer(tend,tstart)
         timediff = tend / rtimes
         write(*,'(I3,9X,f9.4,5X,f8.4)') reps,ceur,timediff
 1111 continue
 9991 print *
C --- stochastic volatility American Model
      print *,'stochastic volatility American Model'
      print *,'reps   reps1       Value       Time(sec.)'
      print *,'-----------------------------------------'
      do 2222 reps = 17, 17
         do 3333 j = 0, reps-1
            tdiv = t*j/float(reps)
            reps1=(INT(reps*tdiv/t)+1)
            CALL timer(tstart, 0.0)
            do 7000 i = 1, times
               call asvcorp2(sz,k,rate,t,v(4),psi,xmu,rho,reps1,reps,
     &              tdiv,cams,
     &              tcomp,tcomms)
 7000       continue
            CALL timer(tend,tstart)
            timediff = tend / rtimes
            write(*,'(I3,5X,I3,5X,f9.4,5X,f8.4)')
     &           reps,(INT(reps*tdiv/t)+1),
     &           cams,timediff
 3333    continue
 2222 continue
C Stop timer
      CALL timer(tf, ti)
      WRITE(UNIT=*,FMT=6004) NUMBER_OF_PROCESSORS(),reps,
     1                       tcomms,tcomp,(tf-tcomms-tcomp),tf
      STOP
 6004 FORMAT('Number of Processors = ',I4/
     1       'Problem size = ',I5/
     2       'Communications = ',F9.3/
     3       'Compute        = ',F9.3/
     4       'Others         = ',F9.3/
     5       'Total time     = ',F9.3)
      END
C======================================================================
      SUBROUTINE asvcorp2(sp,sk,r,t,v,psi,xmu,cor,reps1,reps,div,cval,
     1                    tcomp,tcomms)
      IMPLICIT NONE
      REAL sp,sk,r,t,v,psi,xmu,cor,div,cval
      REAL :: tcomp, tcomms
      INTEGER reps,reps1
      REAL,DIMENSION(1:(2**(reps-12)-1),1:8192) :: ss1,vv1
      REAL,DIMENSION(1:2**(reps-13),1:8192) :: cearly_big
      REAL,DIMENSION(1:8192) :: aatmp1,aatmp2,uux,ddx
      REAL,DIMENSION(1:2**reps1) :: cearly,cback
      REAL u,d,dt,r1,tmp1,tmp2,rx,xreps
      INTEGER i,j,n,nreps1,start,end,repsdiv,reps2,nreps2
      REAL :: tstart,tend

! The HPF statements first declare a 1D abstract processor array
! of size NUMBER_OF_PROCESSORS(), which is specified at run time.
! The 2D arrays are distributed in blocks along their second dimension,
! while the first dimension is collapsed with a *.
! The 1D arrays are distributed in blocks onto the processors.
!HPF$ PROCESSORS P( NUMBER_OF_PROCESSORS() )
!HPF$ DISTRIBUTE (*,BLOCK) ONTO P :: vv1,ss1,cearly_big
!HPF$ DISTRIBUTE (BLOCK) ONTO P:: uux,ddx,cearly,cback,aatmp1,aatmp2
      xreps = float(reps)
      dt = t/xreps
c       --- SET RX (PERIODIC 1+RF) TO RISK FREE RATE + 1
      rx=(r+1.0)**dt
      tmp1 = (xmu-0.5*psi*psi)*dt
      tmp2 = cor*psi*sqrt(dt)
      u = exp(tmp1+tmp2)
      d = exp(tmp1-tmp2)
c       --- create stock price lattice (sl)
      vv1(1,:) = 0.0
      ss1(1,:) = 0.0
      vv1(1,1) = v*v
      ss1(1,1) = sp
      n = 1
      do i = 1,13
         CALL timer(tstart, 0.0)
         aatmp1 = (r-0.5*vv1(1,:))*dt
         aatmp2 = sqrt(vv1(1,:)*dt)
         uux = exp(aatmp1+aatmp2)
         ddx = exp(aatmp1-aatmp2)
         CALL timer(tend, tstart)
         tcomp = tcomp + tend
         CALL timer(tstart, 0.0)
         vv1(1,:) =  u*vv1(1,:) +
     &        eoshift(d*vv1(1,:),dim = 1,shift = -n)
         ss1(1,:) = uux*ss1(1,:) +
     &        eoshift(ddx*ss1(1,:),dim = 1,shift = -n)
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
         if (i.eq.reps1) then
            CALL timer(tstart, 0.0)
            cearly = ss1(1,1:2**i) - sk
            CALL timer(tend, tstart)
            tcomms = tcomms + tend
            CALL timer(tstart, 0.0)
            ss1(1,1:2**i) = ss1(1,1:2**i) - div
            CALL timer(tend, tstart)
            tcomp = tcomp + tend
         endif
         n = 2*n
      enddo
      if (reps.gt.13) then
         reps2 = reps - 13
         nreps1 = 2**(reps2+1) - 1
         repsdiv = reps1 - 14
         start = 2**repsdiv
         end = 2*start - 1
         CALL timer(tstart, 0.0)
         do i = 1, nreps1/2
            aatmp1 = (r-0.5*vv1(i,:))*dt
            aatmp2 = sqrt(vv1(i,:)*dt)
            uux = exp(aatmp1+aatmp2)
            ddx = exp(aatmp1-aatmp2)
            vv1(2*i,:) =  u*vv1(i,:)
            vv1((2*i+1),:) = d*vv1(i,:)
            if (i.ge.start.and.i.le.end) then
               aatmp1 = uux*ss1(i,:)
               aatmp2 = ddx*ss1(i,:)
               cearly_big(2*(i-start)+1,:) = aatmp1 - sk
               cearly_big(2*(i-start+1),:) = aatmp2 - sk
               ss1(2*i,:) =  aatmp1 - div
               ss1((2*i+1),:) = aatmp2 - div
            else
               ss1(2*i,:) = uux*ss1(i,:)
               ss1((2*i+1),:) = ddx*ss1(i,:)
            endif
            ss1(i,:) = 0.0
         enddo
         ss1 = ss1 - sk
         where (ss1.lt.0.0) ss1 = 0.0
         CALL timer(tend, tstart)
         tcomp = tcomp + tend
         if (reps1.le.13) then
            CALL timer(tstart, 0.0)
            do i=nreps1/2+1,nreps1
               ss1(1,:) = ss1(1,:) + ss1(i,:)
            enddo
            ss1(1,:) = ss1(1,:)/((2*rx)**(float(reps-13)))
            n = 1
            CALL timer(tend, tstart)
            tcomp = tcomp + tend
            CALL timer(tstart, 0.0)
            do i=1,13-reps1
               ss1(1,:) = ss1(1,:) +
     &              eoshift(ss1(1,:),dim = 1,shift = n)
               n = 2*n
            enddo
            CALL timer(tend, tstart)
            tcomms = tcomms + tend
            CALL timer(tstart, 0.0)
            nreps2 = 2**(13-reps1)
            cback = ss1(1,1:8192:nreps2)/
     &           ((2*rx)**float(13-reps1))
            where (cearly.gt.cback) cback = cearly
            CALL timer(tend, tstart)
            tcomp = tcomp + tend
            CALL timer(tstart, 0.0)
c       --- calculate call price
            cval = SUM(cback,mask=cback.gt.0.0)/((2*rx)**(float(reps1)))
            CALL timer(tend, tstart)
            tcomms = tcomms + tend
         else
            nreps2 = 2**(reps-reps1)
            cval = 0.0
            do i=1,2*start
               CALL timer(tstart, 0.0)
               aatmp1 = 0.0
               n = nreps1/2 + (i-1)*nreps2
               do j=1,nreps2
                  aatmp1 = aatmp1 +
     &                 ss1(n+j,:)
               enddo
               aatmp1 = aatmp1/((2*rx)**(float(reps-reps1)))
               where (cearly_big(i,:).gt.aatmp1)
     &              aatmp1 = cearly_big(i,:)
               CALL timer(tend, tstart)
               tcomp = tcomp + tend
               CALL timer(tstart, 0.0)
               cval = cval + SUM(aatmp1,mask=aatmp1.gt.0.0)
               CALL timer(tend, tstart)
               tcomms = tcomms + tend
            enddo
c       --- calculate call price
            cval = cval/((2*rx)**(float(reps1)))
         endif
      else
c       --- calculate call price
         ss1 = ss1 - sk
         CALL timer(tstart, 0.0)
         cval =  SUM(ss1(1,:),mask=ss1(1,:).gt.0.0)/((2*rx)**xreps)
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
      endif
      return
      end
C======================================================================
      SUBROUTINE svcorp2(sp,sk,r,t,v,psi,xmu,cor,reps,cval,
     1                    tcomp,tcomms)
      IMPLICIT NONE
      real sp,sk,r,t,v,psi,xmu,cor,cval
      REAL :: tcomp,tcomms
      integer reps
      real,dimension(1:(2**(reps-12)-1),1:8192) :: ss1,vv1,uux,ddx,
     &     aatmp1,aatmp2
      integer,dimension(1:8192) :: index,ud
      real u,d,dt,r1,tmp1,tmp2,rx,xreps
      integer i,reps1,nreps1,n2
      REAL :: tstart,tend
! The HPF statements first declare a 1D abstract processor array
! of size NUMBER_OF_PROCESSORS(), which is specified at run time.
! The 2D arrays are distributed in blocks along their second dimension,
! while the first dimension is collapsed with a *.
! The 1D arrays are distributed in blocks onto the processors.

!HPF$ PROCESSORS P( NUMBER_OF_PROCESSORS() )
!HPF$ DISTRIBUTE (*,block) onto P:: ss1,vv1,uux,ddx,aatmp1,aatmp2
!HPF$ DISTRIBUTE (block) onto P:: index,ud
      CALL timer(tstart, 0.0)
      xreps = float(reps)
      dt = t/xreps
C..HWY.  Corrected syntax from `index = [0:891]'.
      FORALL(i=1:8192) index(i) = i - 1
c       --- SET RX (PERIODIC 1+RF) TO RISK FREE RATE + 1
      rx=(r+1.0)**dt
      tmp1 = (xmu-0.5*psi*psi)*dt
      tmp2 = cor*psi*sqrt(dt)
      u = exp(tmp1+tmp2)
      d = exp(tmp1-tmp2)
c       --- create stock price lattice (sl)
      vv1(1,:) = v*v
      ss1(1,:) = sp
      n2 = 4096
      do i = 1,13
         aatmp1(1,:) = (r-0.5*vv1(1,:))*dt
         aatmp2(1,:) = sqrt(vv1(1,:)*dt)
         uux(1,:) = exp(aatmp1(1,:)+aatmp2(1,:))
         ddx(1,:) = exp(aatmp1(1,:)-aatmp2(1,:))
         ud = index/n2
c         where (mod(ud,2) .ne. 0)
         where (BTEST(ud,POS=0))
            vv1(1,:) =  d*vv1(1,:)
            ss1(1,:) = ddx(1,:)*ss1(1,:)
         elsewhere
            vv1(1,:) =  u*vv1(1,:)
            ss1(1,:) = uux(1,:)*ss1(1,:)
         endwhere
         n2 = n2/2
      enddo
      CALL timer(tend, tstart)
      tcomp = tcomp + tend
      if (reps.gt.13) then
         CALL timer(tstart, 0.0)
         reps1 = reps - 13
         nreps1 = 2**(reps1+1) - 1
         do i = 1, nreps1/2
            aatmp1(i,:) = (r-0.5*vv1(i,:))*dt
            aatmp2(i,:) = sqrt(vv1(i,:)*dt)
            uux(i,:) = exp(aatmp1(i,:)+aatmp2(i,:))
            ddx(i,:) = exp(aatmp1(i,:)-aatmp2(i,:))
            vv1(2*i,:) =  u*vv1(i,:)
            vv1((2*i+1),:) = d*vv1(i,:)
            ss1(2*i,:) = uux(i,:)*ss1(i,:)
            ss1((2*i+1),:) = ddx(i,:)*ss1(i,:)
            ss1(i,:) = 0.0
         enddo
c       --- calculate call price
         ss1 = ss1 - sk
         CALL timer(tend, tstart)
         tcomp = tcomp + tend
         CALL timer(tstart, 0.0)
         cval =  SUM(ss1,mask=ss1.gt.0.0)/((2*rx)**xreps)
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
      else
c       --- calculate call price
         ss1 = ss1 - sk
         CALL timer(tstart, 0.0)
         cval =  SUM(ss1(1,:),mask=ss1(1,:).gt.0.0)/((2*rx)**xreps)
         CALL timer(tend, tstart)
         tcomms = tcomms + tend
      endif
      return
      end

  		subroutine timer(return_time, initial_time)
       real, intent(IN)  :: initial_time
       real, intent(OUT)  :: return_time
       integer finish,rate
        CALL system_clock(COUNT=finish,COUNT_RATE=rate)
        return_time = FLOAT(finish) / FLOAT (rate) -initial_time
        return
       end


