
!A.5 Sweep Over an Unstructured Mesh

! This code represents a simple edge sweep on an unstructured
! mesh. The additional flux at node `i' is the sum of all its
! neighbouring nodes. The edge array `ndeE()' stores the
! adjacency list. The amount of communication generated will
! depend on the locality of access. The cost of the
! communication will be low if the nodes are ordered such that
! most of these accesses are local. This reordering can be
! potentially achieved by use of extrinsic functions to perform
! graph partitioning using the edge array. 

      PROGRAM sweep
      USE HPF_LIBRARY
C     Sweep over an unstructured mesh
      INTEGER, PARAMETER :: ne=3000,nn=500,iter=100
      INTEGER ndeS(ne),ndeE(ne)
      INTEGER i,j
      REAL    flux(nn),econs
      REAL    weight(nn,4),cc(nn)
      REAL    data(ne)

! These HPF statements first create a 1D abstract processor array P,
! onto which the 1D array nseS is distributed in a BLOCK fashion.
! The number of processors used, i.e. the size of P is specified by the
! user at run time.  The array ndeE is aligned with ndeS, insuring that
! they are distributed in the same fashion.  The 2D array weight is
! distributed in a BLOCK manner along its first dimension, and the
! 1D arrays flux & cc, are aligned with this dimension, so the
! distributions are the same.  Note the size of flux and cc must
! equal the size of the first dimension of weight.

!HPF$ PROCESSORS P(NUMBER_OF_PROCESSORS())
!HPF$ DISTRIBUTE ndeS(BLOCK) ONTO P
!HPF$ ALIGN ndeE(:)  WITH  ndeS(:)
!HPF$ ALIGN flux(:)  WITH  weight(:,*)
!HPF$ ALIGN cc(:)    WITH  weight(:,*)
!HPF$ DISTRIBUTE weight(BLOCK,*) ONTO P
C     Initialization
      FORALL(i=1:nn) flux(i) = 0.0
      FORALL(i=1:nn) weight(i,1) = 1.0
      FORALL(i=1:nn) weight(i,2) = 0.5
      FORALL(i=1:nn) weight(i,3) = 0.5
      FORALL(i=1:nn) weight(i,4) = 1.0
      FORALL(i=1:nn) cc(i) = 1.0
      econs = 1.0
      FORALL(i=1:ne)
         ndeS(i) = 1+MOD((i-1),nn)
         ndeE(i) = 1+MOD(ne-(i-1),nn)
      END FORALL
      DO j = 1,iter
C     Weighted Sweep over edges
         FORALL(i=1:ne)
            data(i) = econs*(
     &          ((weight(ndeS(i),2)+weight(ndeS(i),3)+weight(ndeS(i),4))
     &          / weight(ndeS(i),1)
     &          +
     &           (weight(ndeE(i),2)+weight(ndeE(i),3)+weight(ndeE(i),4))
     &          / weight(ndeE(i),1)
     &          ) / 2 + cc(ndeS(i)) )
         END FORALL
         flux = SUM_SCATTER(data,flux,ndeS)
         FORALL(i=1:ne)
            data(i) = econs*(
     &          ((weight(ndeS(i),2)+weight(ndeS(i),3)+weight(ndeS(i),4))
     &          / weight(ndeS(i),1)
     &          +
     &           (weight(ndeE(i),2)+weight(ndeE(i),3)+weight(ndeE(i),4))
     &          / weight(ndeE(i),1)
     &          ) / 2 + cc(ndeE(i)) )
         END FORALL
         flux = SUM_SCATTER(data,flux,ndeE)
      END DO
      DO i=1, nn
         PRINT *,'flux(',i,') = ',flux(i)
      end dO
      END
