/* This random number generator has a fairly small period, and so we
   use rng() instead. Prand is used solely to initialize the seeds
   for rng().
*/

/********
* Prand *	J.W.Flower  April 1985
*********

This random number generator has the same calling sequence and storage 
requirements as the parallel set but works on a sequential machine. To use
it one needs to have defined the following externals which are used here.
*/

long	AAA;	/* multiplier */
long	BBB;	/* summand    */

#include "comext.h"
#include <math.h>
#include <stdio.h>

/*
One then calls pranset with your seed as its argument (once only) and then
succesive random numbers are obtained with the call ranf(). Remember to
declare that ranf() returns a double in every routine which uses it!!
*/

#define	MULT	1103515245
#define	ADD 	12345
#define	MASK	( 0x7fffffff )
#define	TWOTO31	( (double)MASK + 1.0 )

pranset(_s_eed)
long   _s_eed;
{

	AAA = MULT & MASK ;
	BBB = ADD  & MASK ;
	randx = (AAA*_s_eed + BBB) & MASK;
}

double ranf()
/*
	Return a random double in [0, 1.0)
*/
{
	double retvalue;

	retvalue=randx / TWOTO31;
	randx = (AAA*randx + BBB)& MASK;
	return( retvalue );
}


long iranf()
/*
	Return a random long int in [0, 2^31)
*/
{
	randx = (AAA*randx + BBB)& MASK;
	return( randx );
}


/**********
* setseed *    J.W.Flower & the infamous Dr. O  March 1985
***********

Used to set the seed for the parallel random number system -
MUST be called or all processors will execute same sequence! - see
Chapter 12 of *THE BOOK*.  Two options:
  i) Input seed from terminal                        
  ii) Read in seed from standard file "Seed".  
ALSO: at the end of production runs, don't forget to store current
seed in "Seed" by running option "d". 
N.B. You don't need to do this any more, since the seed is dumped at 
the end of the backup file. 
*/

setseed()
{
    FILE *fp ;
    int mode ;
    long seed ;

    printf("Select mode of seed setting : (0=input,1=file) : ") ;
    fflush(stdout);
    scanf("%d",&mode) ;
    printf("%d\n",mode) ;
    fflush(stdout);

    if (mode == 0) {
      printf("Input seed : ");
      fflush(stdout);
      scanf("%ld",&seed);
    }  
    else {
      if ((fp = fopen("Seed","r")) == NULL) {
        printf("Couldn't find Seedfile...assuming 42 !!\n") ;
        fflush(stdout);
        seed = 42L ;
      }
      else {
        fscanf(fp,"%ld",&seed) ;
        fclose(fp) ; 
      }
    }
    pranset(seed) ;

    printf("Seed set to %ld\n",seed);
    fflush(stdout);

    return(0) ;
}

