/*       Ising.c      */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

int Spins[128][128];

double t_rand();
int seed;


/*******************************************************************************

                           PROCEDURE INITIALIZE_LATTICE

*******************************************************************************/


void Initialize_Lattice( Rows, Cols )
int Rows, Cols;

{
   int i, j, Sentinel, Initial_Conf;
   do 
   {
      Sentinel = 0;
      printf ( "\n\nEnter the number next to your choice of spin\n" );
      printf ( "arrangements for the initial lattice.\n\n" );
      printf ( "1 -- All spins up.\n" );
      printf ( "2 -- All spins down.\n" );
      printf ( "3 -- Random configuration.\n\n\n" );
      printf ( "Choice: " );
      scanf ( "%d", &Initial_Conf );

      if ( (Initial_Conf != 1) && (Initial_Conf != 2) && (Initial_Conf != 3) )
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);

   switch (Initial_Conf)
   {

      case 1  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = 1;
           break;

      case 2  : 
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )            
                 Spins[i][j] = -1;
           break;

      case 3  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )
              {
                 if (t_rand() < .5 )
                    Spins[i][j] = 1;
                 else 
                    Spins[i][j] = -1;
              }         
           break;
 
      default  :
           printf("initialization is screwed up!\n");

   }

   printf ( "\n\n" );
}



/*******************************************************************************

                            PROCEDURE INPUT_PARAMETERS

*******************************************************************************/
 

void Input_Parameters( Num_Spins, Rows, Cols, Beta, NTherm, Freq, Num_Sweeps )
int *Num_Spins, *Rows, *Cols;
double *Beta;
int *NTherm, *Freq, *Num_Sweeps;

{
   int i; 
   char Sentinel;

   do   
   {
      Sentinel = 0;
      printf ( "Enter the vertical dimensions of the lattice (0 < Y <= 128): " );
      scanf ( "%d", Rows );

      if ( ( *Rows > 128 ) || ( *Rows <= 0 ) )
      { 
         printf ( "\n\nInput out of range.\n\n" );
         Sentinel = 1;
      }
   } 
   while (Sentinel);

   do
   {
      Sentinel = 0;

      printf ( "\n\nEnter the horizontal dimensions of the lattice (0 < X <= 128): " );
      scanf ( "%d", Cols );

      if ( ( *Cols > 128 ) || ( *Cols <= 0 ) )
      {
            printf ( "\n\nInput out of range.\n\n" );
            Sentinel = 1;
      }
   }
   while (Sentinel);

   *Num_Spins  = (*Rows) * (*Cols);
   printf ( "\n\nLATTICE SIZE :    %d X %d = %d\n", *Rows, *Cols, *Num_Spins );

   do 
   {
      Sentinel = 'n';

      printf ( "Enter the value of Beta: " );
      scanf ( "%lf", Beta );

      printf ( "\n\nEnter the number of thermalization sweeps desired: " );
      scanf ( "%d", NTherm );
      if ( *NTherm < 0 ) *NTherm = 0;

      do
      {
         Sentinel = 'n';        

         printf ( "\n\nEnter the total number of calculation updates desired: " );
         scanf ( "%d", Num_Sweeps );
         if ( *Num_Sweeps <= 0 )   
            {
               printf ( "\n\nInvalid input\n\n" );
               Sentinel = 'y';
            }
      } while (Sentinel == 'y');
     
      do
      { 
         Sentinel = 'n';

         printf ( "\n\nEnter the number of sweeps per calculation update: ");
         scanf ( "%d", Freq );
         if ( *Freq > *Num_Sweeps ) *Freq = *Num_Sweeps;
         if ( *Freq <= 0 ) 
         { 
            printf ( "\n\nInvalid input\n\n" );
            Sentinel = 'y';
         }   
      } while ( Sentinel == 'y' ); 
     
      printf ( "\n\nCurrent value of Beta                       = %10.8lf\n", *Beta );      printf ( "Current number of thermalization sweeps     = %d\n", *NTherm );
      printf ( "Current total number of sweeps              = %d\n", *Num_Sweeps );
      printf ( "Current number of sweeps/calculation update = %d\n\n", *Freq );
      printf ( "Are all values correct (y/n)?" );
      scanf ( "%1s",&Sentinel );
      printf ( "\n\n" );
   } while ( ( Sentinel != 'y' ) && ( Sentinel != 'Y' ) );
}



/*******************************************************************************

                            PROCEDURE METROPOLIS_UPDATE

*******************************************************************************/


void Metropolis_Update ( Beta, Rows, Cols, Accept )
double Beta;
int Rows, Cols, *Accept;

{
   int i, j, ipos, ineg, jpos, jneg;
   int Old_Spin, New_Spin, Spin_Sum;
   double Old_Energy, New_Energy, Energy_Diff;

   *Accept = 0;

   /* printf("Top of Metro:: Rows = %d, Cols = %d \n",Rows, Cols); */
   for ( i = 0; i < Rows; i++ )
      {
         if (i < Rows-1) 
            (ipos = i+1);
         else 
            (ipos = 0);
 
         if (i > 0)
            (ineg = i-1);
         else 
            (ineg = Rows-1);
         
         for ( j = 0; j < Cols; j++ ) {
               if (j < Cols-1)
                  (jpos = j+1);
               else 
                  (jpos = 0);
               
               if (j > 0) 
                  (jneg = j-1);
               else 
                  (jneg = Cols-1);

               Old_Spin = Spins[i][j];
               New_Spin = -Old_Spin;

               Spin_Sum = Spins[i][jpos] + Spins[ipos][j] + Spins[i][jneg] + Spins[ineg][j];

               Old_Energy = -Beta * Old_Spin * Spin_Sum;
               New_Energy = -Beta * New_Spin * Spin_Sum;
               Energy_Diff = New_Energy - Old_Energy;

               if ( (Energy_Diff <= 0.0) || ( exp(-Energy_Diff) > t_rand() ) )
               { 
                  Spins[i][j] = New_Spin;
                  *Accept = *Accept+1;
               }
            }
      }
}



/*******************************************************************************

                            PROCEDURE CALCULATE

*******************************************************************************/


void Calculate ( Beta, Rows, Cols, Energy, Magnet )
double Beta;
int Rows, Cols;
double *Energy, *Magnet;

{
   int i, ineg, j, jneg, Sum_Spins;

   *Magnet = 0.0;
   *Energy = 0.0;
   Sum_Spins = 0;

   for ( i = 0; i < Rows; i++ ) {  
      if (i > 0)
         (ineg = i-1);
      else
         (ineg = Rows-1);
         
      for ( j = 0; j < Cols; j++ ) {               
         if (j > 0)
            (jneg = j-1);
         else
            (jneg = Cols-1);

         Sum_Spins += Spins[i][j] * (Spins[ineg][j] + Spins[i][jneg]);
         *Magnet += (double) Spins[i][j];

      } 

   }

   *Energy = -1.0 * (double) Sum_Spins;
}



/*******************************************************************************
                                
                                MAIN PROGRAM

*******************************************************************************/

void main()
{
   int Sweep;
   int Lattice;  
   int NTherm;
   int Rows, Cols;     
   int Num_Spins;      
   int Freq;          
   int Num_Sweeps;     
   int Accept; 
   
   extern int Spins[128][128]; 
     
   double Beta;          
   double Energy;       
   double Magnet;         
   double Total_Energy = 0;  
   double Total_Magnet = 0;   

   void Initialize_Lattice (); 
 /*     int *Num_Spins,
                             int *Rows, int *Cols );
 */

   void Input_Parameters ();
 /*  double *Beta,
                           int *NTherm,
                           int *Freq,
                           int *Num_Sweeps );
 */
  
   void Metropolis_Update ();
 /*      double Beta,
                            int Rows, int Cols,
                            int *Accept );
 */  

   void Calculate ();
 /*       double Beta,
                    int Rows, int Cols,
                    double *Energy, double *Magnet );
 */

   FILE *Outfile;

   Outfile = fopen ( "info", "w+");

   seed = 734637;
   srand(seed);
   
   Input_Parameters( &Num_Spins, &Rows, &Cols, 
		     &Beta, &NTherm, &Freq, &Num_Sweeps );

   Initialize_Lattice( Rows, Cols );

   fprintf ( Outfile, "THERMALIZATION SWEEPS :\n\n" );
   fprintf ( Outfile, "Acceptance \n" );
   fprintf ( Outfile, "------------\n\n\n" );

   for ( Sweep = 0; Sweep < NTherm; Sweep++ )
   {
         Metropolis_Update( Beta, Rows, Cols, &Accept );
         fprintf ( Outfile, "%f\n", ( (float)Accept/(float)Num_Spins ) ); 

   }

   fprintf ( Outfile, "\n\n\n" );
   fprintf ( Outfile, "BEGIN CALCULATIONS :\n\n" );
   fprintf ( Outfile, "LATTICE SIZE :    %d X %d = %d\n", Rows, Cols, Num_Spins );
   fprintf ( Outfile, "BETA         :    %10.8lf\n\n\n\n", Beta );
   fprintf ( Outfile, "CALCULATION    ACCEPTANCE        ENERGY       MAGNETIZATION\n" );   
   fprintf ( Outfile, "-----------    ----------        ------       -------------\n" );

   for ( Sweep = 0; Sweep < Num_Sweeps; Sweep++ )
   {
      for ( Lattice = 0; Lattice < Freq; Lattice++ )
         Metropolis_Update( Beta, Rows, Cols, &Accept );
       
      Calculate( Beta, Rows, Cols, &Energy, &Magnet );

      Energy /= (double)Num_Spins;
      Magnet /= (double)Num_Spins;

      Total_Energy += Energy;
      Total_Magnet += Magnet;

      fprintf ( Outfile, "%6d \t\t %6.4f \t %10.8lf \t %10.8lf\n", 
              Sweep+1,  (float)Accept/(float)Num_Spins, 
              Total_Energy/(double)(Sweep+1),Total_Magnet/(double)(Sweep+1) ); 
   }

   fclose ( Outfile );

}       
           

double t_rand()
{
 double temp;

 temp = (double)rand()/(double) RAND_MAX; 
 if (temp == 1.0) temp = 0.0;

 if ( (temp >1.0) || (temp < 0 ) ) {
     perror("ERROR: t_rand is bad");
     printf("trand = %lf \n",temp);
     exit(-2);
 }

 return(temp);

}

