/******************************************************
*         PARALLEL RANDOM NUMBER GENERATOR            *
*       USING SHIFT-REGISTER SEQUENCE METHOD          *
*                                                     *
*                      by                             *
*                                                     *
*                Ting-Wai Chiu        Sept,1987.      *
*                                                     *
*    Changed by Paul Coddington so that it works,     *
*    Oct, 1990.                                       *
*    Tidied up a lot, May 1992.                       *
*                                                     *
*    The CUBIX flag allows it to work on parallel     *
*    machines under Express.                          *
*                                                     *
******************************************************/

#include "comext.h"
#include <math.h>
#include <stdio.h>

#define PT0_INIT    0x00dd    /* Initialize pointers for TWC rng    */
#define PT147_INIT  0x004a    /* (instead of reading from pts file) */

#define A 3141592621 
#define MASK ( 0x7fffffff )

int seeds[250];
int pt0,pt147;
double twoto31 = 2147483648.0;


/* Generate the initial seeds and pointers */
void init_seeds()
{
   long iranf();
   long tmp,tmp1;
   short int i,j;

   tmp1 = iranf();
   tmp = tmp1;
   for (i=249;i>=0;i--) {
      tmp1 = iranf();
      tmp = (tmp * A + tmp1) & MASK;
      seeds[i] = tmp^tmp1;
   }
   pt0 = PT0_INIT;
   pt147 = PT147_INIT;
}


/* Read in backed up seeds and pointers */
void read_seeds(seedname,ptsname)
char seedname[128], ptsname[128];
{
   FILE *f1,*f2,*fopen();

   f1 = fopen(seedname,"r");
#ifdef CUBIX
   fmulti(f1);
#endif CUBIX
   fread(seeds,sizeof(long),250,f1);
   fclose(f1);
   f2 = fopen(ptsname,"r");
#ifdef CUBIX
   fmulti(f2);
#endif CUBIX
   fscanf(f2,"%04x %04x",&pt0,&pt147);
   fclose(f2);
}


/* Back up seeds and pointers */
void write_seeds(seedname,ptsname)
char seedname[128], ptsname[128];
{
   FILE *f1,*f2,*fopen();

   f1 = fopen(seedname,"w") ;
#ifdef CUBIX
   fmulti(f1);
#endif CUBIX
   fwrite(seeds,sizeof(long),250,f1);
   fclose(f1);
   f2 = fopen(ptsname,"w");
#ifdef CUBIX
   fmulti(f2);
#endif CUBIX
   fprintf(f2,"%04x %04x\n",pt0,pt147);
   fclose(f2);
}


double pranf()
{
   int tmp ;
   int t0,t147 ;

   t0 = pt0 ;
   t147 = pt147 ;
   tmp = seeds[t147] ;
   /* Addition is better than OR */
   /* seeds[t0] ^= tmp ; */
   seeds[t0] = (seeds[t0] + tmp) & MASK ;
   if (--t0 < 0) {
      t0 = 249;
   }
   if (--t147 < 0) {
      t147 = 249;
   }
   pt0 = t0 ;
   pt147 = t147 ;
   
   return((double)tmp / twoto31) ;
}


double normal()
{
   double x ;
   int tmp ;
   int t0,t147 ;
   char loop ;
 
   t0 = pt0 ;
   t147 = pt147 ;
   x = 0 ;
   for (loop=12;loop!=0;loop--) {
      tmp = seeds[t147] ;
      /* Addition is better than OR */
      /* seeds[t0] ^= tmp ; */
      seeds[t0] = (seeds[t0] + tmp) & MASK ;
      x += tmp ;
      if (--t0 < 0) {
         t0 = 249 ;
      }
      if (--t147 < 0) {
         t147 = 249 ;
      }
   }
   pt0 = t0 ;
   pt147 = t147 ;
   return( ldexp( (double) x, (int) -31) - 6) ;
}


int ipranf(imax)
int imax ;
{
   int tmp ;
   int t0,t147 ;

   t0 = pt0 ;
   t147 = pt147 ;
   tmp = seeds[t147] ;
   /* Addition is better than OR */
   /* seeds[t0] ^= tmp ; */
   seeds[t0] = (seeds[t0] + tmp) & MASK ;
   if (--t0 < 0) {
      t0 = 249 ;
   }
   if (--t147 < 0) {
      t147 = 249 ;
   }
   pt0 = t0 ;
   pt147 = t147 ;
   return(tmp%imax) ;
}

