/*+---------------------------------------------------------------------+
  |                         File: hist.c                                |
  +---------------------------------------------------------------------+*/

#include "hist.h"

#define	MAXA	5000
#define	MAXB	75

static	int	hist[MAXA][MAXB];

static	struct {	
	int	ibeta;	/* pointer to current beta in the histogram 	*/
	int	ibin;	/* pointer to current bin (magnetization)	*/
	int	nbins;	/* number of bins				*/
	} h;
	
static	void	fheader();
	
void	init_hist()
{
	int	i,j,sz;
	struct	Stats	s;
		
	sz = pRows()*pCols();
	h.ibeta = indexB();
	h.nbins =  sz + 1;
	assert(numBs()<MAXB && h.nbins<MAXA);
	
	for (i=0;i<numBs();i++)
	   for (j=0;j<h.nbins;j++) 
	      hist[i][j] = 0;
	      
	s = calculate();
	h.ibin = (s.Magnet + sz)/2;
}


void	update_hist(old,new)
	int	old;
	int	new;
{
	struct	Stats s;
	
	h.ibeta = indexB();
	if (old!=new) 
	   h.ibin += (new==1) ? 1 : -1;
	assert((h.ibin>=0)&&(h.ibin<h.nbins));
	hist[h.ibeta][h.ibin]++;
	
/*	printf("%d %d %d\n", h.ibeta,h.ibin,hist[h.ibin][h.ibeta]);*/
}


void	write_hist()
{
	int	i,j;
	FILE	*fp;
	
	if (!pHist()) return;
	fp = fopen("hist.dat","w");
	if (fp==NULL) {
	   printf("Can't open file [hist.dat].\n");
	   exit(-1);
	}
	fheader(fp);
	for (i=0;i<numBs();i++) {
	   for (j=0;j<h.nbins;j++)
	      fprintf(fp,"%d \t%d \t%d\n",i,j,hist[i][j]);	
	   fprintf(fp,"\n");
	}
}


static  void	fheader(fpointer)
	FILE	*fpointer;
{
	assert(fpointer!=NULL);
	fprintf(fpointer,"#             size: %d\n",pRows());
	fprintf(fpointer,"# number of sweeps: %d\n",pNSweeps());
	fprintf(fpointer,"#            alpha: %f\n",(float)pAlpha());
	fprintf(fpointer,"#            model: %s\n",
		pIsing() ? "ISING" : "SPIN GLASS");
}
