/*+---------------------------------------------------------------------+
  |			File: beta.c					|
  +---------------------------------------------------------------------+*/
 
#include "beta.h"


static	struct	Beta {
	double	betas[MAXBETAS];
	int	cur;
	int	last;
	int	dir;
	int	nbetas;
	int	visit[MAXBETAS];
	} b;

static	double	look_up();
static	double	look_down();
static	double	go_up();
static	double	go_down();


void	init_betas()
/*+---------------------------------------------------------------------+
  |	Initializes list of betas according to an exponential schedule.	|
  +---------------------------------------------------------------------+*/
{
	int	i;
	double	beta0,betaf;
	double	alpha;
	
	alpha = pAlpha();
	assert(alpha>0.0);
	beta0 = pBeta0();
	betaf = pBetaF();			
	b.cur = b.last = 0;
	b.nbetas = 1;
	b.dir = DOWN;
	b.betas[0] = beta0;
	for (i=1; i < MAXBETAS; i++) {
	   b.visit[i]=0;
	   b.betas[i] = b.betas[i-1]/alpha;
	   if (b.betas[i]<betaf) b.nbetas++;
	}
}


double 	tryB()
/*+---------------------------------------------------------------------+
  | 	Returns a proposed new beta according to the following     	|
  |	procedure:							|
  |		o generate random number, x				|
  |		o if m==0 return beta(m+1)				|
  |		o if m==LAST retunr beta(m-1)				|
  |		o if x < 0.5 return beta(m+1)				|
  |		  else return beta(m-1)                 		|  
  +---------------------------------------------------------------------+*/
{
	double	r;
	
	b.visit[b.cur]++;
	if (pBeta0()==pBetaF()) return b.betas[0]; /*curB();*/
	r = kiss();
	if (r<0.5)
	   return (curB()>=thisB(numBs()-1)) ? look_up() : look_down();
	else
	   return (curB()==pBeta0()) ? look_down() : look_up();
}


void	updateB()
{
	if (pBeta0()==pBetaF()) return;
	if (b.dir==UP)
	   go_up();
	else
	   go_down();
}


int	numBs()
{	return b.nbetas;}


int	indexB()
{	return b.cur;}


double	thisB(i)
{
	assert(i>=0 && i<MAXBETAS);
	return b.betas[i];
}


int	lastBi()
{	return b.last;}


int	visitsB(i)
{
	assert(i>=0 && i<MAXBETAS);
	return b.visit[i];
}

	
int	tryDir()
{	return b.dir;}


double	deltaB()
{
	if (tryDir()==UP)
	   return -(b.betas[b.cur-1] - curB());
	return b.betas[b.cur+1] - curB();
}

  
double	curB()
{	return b.betas[b.cur];}


static	double 	look_up()
{
	b.dir = UP;
	return b.betas[b.cur-1];
}


static	double	look_down()
{
	b.dir = DOWN;
	return b.betas[b.cur+1];
}


static	double	go_down()
{
	assert(b.cur<MAXBETAS-1);
	if (++b.cur>b.last)
	   ++b.last;
	return curB();
}


static 	double	go_up()
{
	assert(b.cur>0);
	return b.betas[--b.cur];
}


void	dispB()
{
	int	i;
	
	printf("\n\n");
	printf("\t\t\t beta           # of visits \n\n");
	for (i=0; i<=b.last; i++)
	   printf("\t\t %15.5f \t%d\n",b.betas[i],b.visit[i]);
}


void	btest()
{
	int 	i;
	FILE	*fp;
	
	init_params();
	init_betas();
	
	fp = fopen("beta.dat","w");
	assert(fp!=NULL);	
	printf("\n\t\tGenerating file [beta.dat]...\n");
	for (i=0;i<pNSweeps();i++) {
	   fprintf(fp,"\t [%3d] %4d \t %f\n", i,b.cur,tryB());
	   updateB();
	}
	dispB();
	exit(-1);
}
