/*+---------------------------------------------------------------------+
  |                             File: params.c                          |
  +---------------------------------------------------------------------+*/

#include "params.h"

#define	TRUE		1
#define	FALSE		0
#define	ISING		0
#define	SPINGLASS	1

static	void	defaults();
static	void	new_params();
static	void	switches();
static 	void	boom();

static	struct	Switches {
	int	hist;
	int	zflag;
	int	debug;
	int	model;
	} sw;

static	struct  Params {
        int     rows;
        int     cols;
        int     tsweeps;
        int     nsweeps;
        long	seed;
        double  alpha;
        double  beta0;
        double 	betaf;
        } p;


void 	init_params(argc,argv)
	int	argc;
	char	*argv[];
{
	char	yn;

	defaults();
	switches(argc,argv);	
  	disp_params();
	printf("\n\n\t  Change default parameters? [y,n] ");
	fscanf(stdin,"%s",&yn);
	if (yn=='n')
	   return;
	new_params();
}


static	void	defaults()
{
	p.rows = 10;
	p.cols = 10;
	p.tsweeps = 60;
	p.nsweeps = 15;
	p.alpha = 0.85;
	p.beta0 = 0.08;
	p.betaf = 1.12;
	p.seed = 87654321;
	sw.model = ISING;
	sw.zflag = FALSE;
	sw.debug = FALSE;
  	setkiss(p.seed); 
}


static	void	switches(argc,argv)
/*+---------------------------------------------------------------------+
  |	Extracts command line switches.					|
  |	See README for details.						|
  +---------------------------------------------------------------------+*/
	int	argc;
	char	*argv[];
{
	int	i;
	
	if (argc==1) return;
	if (argc>5) boom();
	for (i=1;i<argc;i++) {
	   if (strcmp(argv[i],"-z")==0)
	      sw.zflag = TRUE;
	   if (strcmp(argv[i],"-d")==0) 
	      sw.debug = TRUE;
	   if (strcmp(argv[i],"-h")==0)
	      sw.hist = TRUE;
	   if (strcmp(argv[i],"-sg")==0)
	      sw.model = SPINGLASS;
	   if (!(sw.debug) && !(sw.zflag) && !(sw.hist) && !(sw.model)) boom();
	}
}

/* split this into init_temper_params and init_model_params */

static	void	new_params()
/*+---------------------------------------------------------------------+
  |	Prompts user for new runtime parameters.			|
  |	See README for details.						|
  +---------------------------------------------------------------------+*/
{
	float	A,B;
	
	printf("\n\n");
	printf("\t\t                 Enter:\n\n");
	printf("\t\t          Lattice size --> ");
	fscanf(stdin,"%d",&p.rows);
	p.cols=p.rows;
	printf("\t\t       Update attempts --> ");
	fscanf(stdin,"%d",&p.nsweeps);
	printf("\t\t                 Alpha --> ");
	fscanf(stdin,"%f",&A);
	p.alpha = (double)A;
	printf("\t\t          Initial Beta --> ");
	fscanf(stdin,"%f",&B);
	p.beta0 = (double)B;
	printf("\t\t            Final Beta --> ");
	fscanf(stdin,"%f",&B);
	p.betaf = (double)B;
	disp_params();
}


void 	disp_params()
{
	printf("\n\n");
	printf("\t\t Simulation parameters:\n\n");
	printf("\t\t          Lattice size --> %d x %d\n", p.rows, p.cols);
	printf("\t\t                 Model --> %s\n",
		(sw.model==ISING) ? "ISING" : "SPIN GLASS");
	printf("\t\t Thermalization sweeps --> %d\n", p.tsweeps);
	printf("\t\t       Update attempts --> %d\n", p.nsweeps);
	printf("\t\t                  Seed --> %d\n", p.seed);
	printf("\t\t            Final beta --> %f\n", p.betaf);
	printf("\t\t          Initial beta --> %f\n", p.beta0);
	printf("\t\t                 Alpha --> %f\n", p.alpha);
	printf("\t\t                  Zoom --> %s\n", 
		(sw.zflag==TRUE) ? "ON" : "OFF");
	printf("\t\t                 Debug --> %s\n", 
		(sw.debug==TRUE) ? "ON" : "OFF");
	if (sw.hist==TRUE)
	   printf("\t\t       Generating file     [hist.dat]\n");
}


static  void	boom()
{
	printf("\n\nUsage: temper [-z] [-d] [-h] [-sg]\n");
	exit(-1);
}


int	pRows()
{	return p.rows;}


int	pCols()
{	return p.cols;}

double	pAlpha()
{	return p.alpha;}


double	pBeta0()
{	return p.beta0;}

double	pBetaF()
{	return p.betaf;}


int	pNSweeps()
{	return p.nsweeps;}


int	pTSweeps()
{	return p.tsweeps;}

int	pHist()
{	return sw.hist;}

int	pZoom()
{	return sw.zflag;}

int	pDebug()
{	return 	sw.debug;}

int	pIsing()
{
	if (sw.model==ISING)
	   return TRUE;
	return FALSE;
}
