/*+---------------------------------------------------------------------+
  |                     File: ranmar.h                                  |
  +---------------------------------------------------------------------+*/

#include "ranmar.h"

static	double	ranmar();
static	int	flag;

void 	ktest()
/*+---------------------------------------------------------------------+
  | 	Generates and prints to stdout 100 random numbers.		|
  +---------------------------------------------------------------------+*/
{
	double 	x;
	int	i;
	
	printf("\n\n\t Random number generator test:\n\n");
	setkiss(87654321);
	for (i=0; i<100; i++) 
	   printf("\t [%3d] \t %f\n",i,kiss());
	exit(0);
}    


void	setkiss(seed)
/*+---------------------------------------------------------------------+
  | 	Seed random number generator.					|
  +---------------------------------------------------------------------+*/
	long	seed;
{
	if (flag) {
	   printf("\t\t Random number generator already seeded.\n");
	   return;
	}
	assert((seed>=0) && (seed<=900000000));	
	ranmar(seed);
	flag = 1;
}


double	kiss()
/*+---------------------------------------------------------------------+
  | 	Returns random number (0.0<=1.0).				|
  +---------------------------------------------------------------------+*/
{
	double r;
	
	assert(flag);
	r = ranmar(0);
	assert((r>=0.0) && (r<=1.0));
	return r;
}


/*+---------------------------------------------------------------------+
  This is a C language implementation of the universal random number generator
  proposed by G. Marsaglia and A. Zaman and translated from F. James' version.
  Parallel initialization follows a suggestion of S. Plimpton.
  D. Levine  Argonne National Laboratory   10/26/92

  F. James
  A review of pseudorandom number generators
  Computer Physics Communication
  60 (1990) 329-344

  G. Marsaglia, A. Zaman, W. Tseng
  Stat Prob. Letter
  9 (1990) 35.

  G. Marsaglia, A. Zaman
  FSU-SCRI-87-50

  This algorithm is a combination of a lagged Fibonacci and arithmetic sequence
  (F. James) generator with period of 2^144.  It provides 32-bit floating point
  numbers in the range from zero to one.  It is claimed to be portable and
  provides bit-identical results on all machines with at least 24-bit
  mantissas.

  ranmar should be initialized with a 32-bit integer seed such that
  0 <= seed <= 900,000,000.  Each of these 900,000,000 values gives rise
  to an independent sequence of ~ 10^30.  
 +---------------------------------------------------------------------+*/
  
static  double	ranmar(newseed)
	int 	newseed;
{
    /* initialization variables */
	int ij, kl, i, j, k, l, m, ii, jj;
	float s, t;

    /* random number variables */
	static int seed=1;      /* default seed if none specified */
	static int i96, j96;
	static float u[97], uni, c, cd, cm;

    /* initialization */    
    	if (newseed!=0) {
           seed = newseed % 900000000;
           ij   = seed / 30082;
           kl   = seed - 30082 * ij;
           i    = ((ij/177) % 177) + 2;
           j    = ( ij      % 177) + 2;
           k    = ((kl/169) % 178) + 1;
           l    = ( kl      % 169);
           for (ii=0; ii<97; ii++) {
              s = 0.0;
              t = 0.5;
              for (jj=0; jj<24; jj++) {
                 m = ( ((i*j) % 179) * k ) % 179;
                 i = j;
                 j = k;
                 k = m;
                 l = ((53*l) + 1) % 169;
                 if (((l*m) % 64) >= 32)
                    s += t;
                 t *= .5;
              }
              u[ii] = s;
           }
        c   = 362436.  /16777216.;
        cd  = 7654321. /16777216.;
        cm  = 16777213./16777216.;
        i96 = 96;
        j96 = 32;            
	}

    /* random number generation */
    	uni = u[i96] - u[j96]; 
    	if ( uni < 0. ) uni += 1.0;
    	u[i96] = uni;
    	i96--;
    	if ( i96 < 0  ) i96 = 96;
    	j96--;
    	if ( j96 < 0  ) j96 = 96;
    	c   -= cd;
    	if ( c   < 0. ) c += cm;
    	uni -= c;
    	if ( uni < 0. ) uni += 1.0;
    	return(uni);
}


