

/***********************************************************************
 *                                                                     *
 *          This software copyrighted by Caltech, 1988.                *
 *                                                                     *
 *          Author:  Ed Felten                                         *
 *                                                                     *
 ***********************************************************************/


#include "opt.h"

#define MAX_K   MAXCITIES

extern int graphics;


void monte_carlo(npts,tour,dist,coord,num_sweeps,start_temp,end_temp)
    int     npts;
    int     tour[MAXCITIES];
    int     dist[MAXCITIES][MAXCITIES];
    int     coord[MAXCITIES][2];
    int     num_sweeps;
    double  start_temp,end_temp;
{
    int     i,sweep,delta,k_change;
    int     save_tour[MAXCITIES];
    double  temperature,beta;

    double  log(),exp(),my_exp(),ranf();

    k_change = npts/5;

    for(sweep=0; sweep<num_sweeps; ++sweep){
	temperature = start_temp*exp(sweep*log(end_temp/start_temp)/num_sweeps);
	beta = 1.0/temperature;

	for(i=0; i<npts; ++i){
	    save_tour[i] = tour[i];
	}

        scramble(npts,tour,k_change);
        three_opt(npts,tour,dist,coord,PLOT_NEVER);
        delta = tour_length(npts,tour,dist)-tour_length(npts,save_tour,dist);
        if(delta == 0){
            k_change += npts/20;
        }
        else{
            k_change = k_change*3/4;
        }
        if( (delta > 0) && (ranf() >= my_exp(-delta*beta)) ){
            printf("T=%f  length=%d  REJECTED\n",
                      temperature,tour_length(npts,tour,dist));
	    for(i=0; i<npts; ++i){
		tour[i] = save_tour[i];
	    }
	}
        else{
            printf("T=%f  length=%d\n",temperature,tour_length(npts,tour,dist));
            if(graphics){
                plot(npts,tour,coord);
	    }
        }
    }
}



scramble(npts,tour,k_change)
    int     npts;
    int     tour[MAXCITIES];
    int     k_change;
{
    int     start[MAX_K],end[MAX_K],reverse[MAX_K];
    int     temp_tour[MAXCITIES];
    int     i,j,temp,new,old;

    if(k_change >= npts){
        random_tour(npts,tour);
        return;
    }

try_again:
    /*    generate random links to break  */
    for(i=0; i<k_change; ++i){
	start[i] = random_int(npts);
	reverse[i] = random_int(2);
    }

    /*   now sort them  */
    for(i=0; i<k_change; ++i){
	for(j=k_change-1; j>i; --j){
	    if( start[j] < start[j-1] ){
		temp = start[j];
		start[j] = start[j-1];
		start[j-1] = temp;
	    }
	}
    }

    /*  look for duplicates  */
    for(i=0; i<k_change-1; ++i){
	if(start[i] >= start[i+1]){
	    start[i+1] = start[i]+1;
            if(start[i+1] >= npts){
                goto try_again;
            }
	}
    }

    /*  fill in the ends  */
    for(i=0; i<k_change-1; ++i){
	end[i] = start[i+1]-1;
    }
    end[k_change-1] = (start[0]+npts-1)%npts; 

    /*  mix 'em up  */
    for(i=k_change-1; i>0; --i){
	new = random_int(i+1);
	temp = start[new];
	start[new] = start[i];
	start[i] = temp;
	temp = end[new];
	end[new] = end[i];
	end[i] = temp;
    }

    /*  now play 'em  */
    for(i=0; i<npts; ++i){
       temp_tour[i] = tour[i];
    }
    new = 0;
    for(i=0; i<k_change; ++i){
	if(reverse[i]){
	    for(old=end[i]; old!=start[i]; old=(old+npts-1)%npts){
		tour[new++] = temp_tour[old];
	    }
	    tour[new++] = temp_tour[start[i]];
	}
	else{
	    for(old=start[i]; old!=end[i]; old=(old+1)%npts){
		tour[new++] = temp_tour[old];
	    }
	    tour[new++] = temp_tour[end[i]];
	}
    }
}


/*
 *	fast exponential
 */

#define RESOLUTION 1000
#define CUTOFF     10


double fine_scale[RESOLUTION];
double coarse_scale[CUTOFF];


init_my_exp()
{
    int     i;

    for(i=0; i<CUTOFF; ++i){
	coarse_scale[i] = exp((double)(-i));
    }
    for(i=0; i<RESOLUTION; ++i){
	fine_scale[i] = exp(((double)(-i))/((double)RESOLUTION));
    }
}


double my_exp(arg)
    double  arg;
{
    int     coarse,fine;

    arg = -arg;
    if(arg >= CUTOFF){
	return(0.0);
    }

    coarse = (int)arg;
    fine = (int)(RESOLUTION*(arg-coarse));
    return(coarse_scale[coarse]*fine_scale[fine]);
}
