/***************************************************/
/* Student : Chien-Hsing Wu                        */
/* S. S. No : 105-76-3704                          */
/* File name : anneal.c                            */
/* Compile Instruction : cc anneal.c -o anneal -lm */
/* Run instruction : anneal Infofile	           */
/* This program calcultate delta gm's for temper.c.*/
/* The Beta schedule is controled by a factor in   */
/* the following way: Beta(m+1) = Beta(m)/factor.  */
/* Delta gm's are estimated by disecting delta Beta*/
/* into n subintervals, and calculate avge Energy  */
/* for each Beta's.				   */
/***************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAXROWS 256
#define MAXCOLS 256
#define MAXBETAS 10

int Spins[MAXROWS][MAXCOLS];    /* spins */
int J[MAXROWS][MAXCOLS][2];     /* couplings */

double Beta_array[MAXBETAS];
double Avge_Energy[MAXBETAS];

int Rows, Cols, Num_Spins;

float Rand();


main(argc, argv)
int  argc;
char  *argv[];
{
   int Sweep;
   int Lattice;  
   int NTherm;
   int Cur_Bet_Ptr;
   int Num_Betas;
   int Initial_Conf;     
   int Freq;          
   int Num_Sweeps;     
   int Num_Measure;     
   int Measurements;     
   int n;
   int Model;
   int Seed;
     
   double Beta;          
   double Energy;       
   double Magnet;         
   double Accept; 
   double Total_Energy = 0.0;  
   double Total_Magnet = 0.0;   
   double Total_Accept = 0.0;   
   double Tot_Energy_Sq = 0.0;  
   double Tot_Magnet_Sq = 0.0;   
   double Tot_Accept_Sq = 0.0;   
   double Error_Energy;  
   double Error_Magnet;   
   double Error_Accept; 
   double Min_Energy = 1000.0;  
   double Specific_Heat; 
   double G_Diff; 
   double Beta_Diff; 
   double Delta_Cm;
   double Beta_Start;
   double Beta_End;
   double Beta_Factor;

   void Input_Parameters ();
   void Metropolis_Update ();
   void Initialize_Lattice (); 
   void Calculate ();

   FILE *Outfile;


  if (argc != 2)
  {
    printf("Usage : anneal Infofile\n");
    exit(1);
  }

  if ((Outfile = fopen(argv[1], "w")) == NULL)
  {
    printf("\nError Opening the Infofile : %s\n", argv[1]);
    exit(2);
  }

  Input_Parameters( &Num_Spins, &Rows, &Cols, &Initial_Conf,
		    &NTherm, &Freq, &Num_Sweeps,  &Beta_Start, &Beta_End,
		    &Beta_Factor, &n,  &Seed, &Model );

  /* Initialize random numbers */
  Rand_Init(Seed);

  Initialize_Lattice ( Rows, Cols, Initial_Conf, Model );

  fprintf ( Outfile, "Beta\t\tG_Diff\n" );

  Beta = Beta_Start;
  Total_Energy = 0.0;
  Total_Magnet = 0.0;
  Total_Accept = 0.0;

  Tot_Energy_Sq = 0.0;
  Tot_Magnet_Sq = 0.0;
  Tot_Accept_Sq = 0.0;

  for ( Sweep = 0; Sweep < Num_Sweeps; Sweep++ )
  {
    for ( Lattice = 0; Lattice < Freq; Lattice++ )
    {
      Metropolis_Update (  Beta, Rows, Cols, &Accept );
    }
 
    Calculate ( Rows, Cols, &Energy, &Magnet );

    if (Energy < Min_Energy) Min_Energy = Energy;

    if ( Sweep > NTherm )
    {
       Total_Energy += Energy;
       Total_Magnet += Magnet;
       Total_Accept += Accept;

       Tot_Energy_Sq += Energy * Energy;
       Tot_Magnet_Sq += Magnet * Magnet;
       Tot_Accept_Sq += Accept * Accept;
    }
  }

  Measurements = Num_Sweeps - NTherm;
  Total_Energy /= Measurements;
  Total_Magnet /= Measurements;
  Total_Accept /= Measurements;

  Tot_Energy_Sq /= Measurements;
  Tot_Magnet_Sq /= Measurements;
  Tot_Accept_Sq /= Measurements;

  Error_Energy = Tot_Energy_Sq - (Total_Energy * Total_Energy);
  Error_Magnet = Tot_Magnet_Sq - (Total_Magnet * Total_Magnet);
  Error_Accept = Tot_Accept_Sq - (Total_Accept * Total_Accept);

  Error_Energy /= Num_Spins;
  Error_Magnet /= Num_Spins;
  Error_Accept /= Num_Spins;

  printf("%.8f\t%.8f\t%.8f\n", Beta, Total_Energy, Error_Energy );
  fflush(stdout);

  do
  {

    Beta_array[0] = Beta;
    Avge_Energy[0] = Total_Energy;
    G_Diff = Total_Energy;

    Beta_Diff = Beta/Beta_Factor;
    if (Beta_Diff < Beta_End)  Beta_Diff -= Beta;
    else		       Beta_Diff = Beta_End - Beta;
    Beta_Diff /= (double)n;

    for ( Cur_Bet_Ptr = 1; Cur_Bet_Ptr <= n; Cur_Bet_Ptr++ )
    {
       Beta_array[Cur_Bet_Ptr] = Beta + Beta_Diff * Cur_Bet_Ptr;

       Total_Energy = 0.0;
       Total_Magnet = 0.0;
       Total_Accept = 0.0;
 
       Tot_Energy_Sq = 0.0;
       Tot_Magnet_Sq = 0.0;
       Tot_Accept_Sq = 0.0;

       for ( Sweep = 0; Sweep < Num_Sweeps; Sweep++ )
       {
         for ( Lattice = 0; Lattice < Freq; Lattice++ )
         {
            Metropolis_Update (  Beta_array[Cur_Bet_Ptr], Rows, Cols, &Accept );
         }
 
         Calculate ( Rows, Cols, &Energy, &Magnet );

         if (Energy < Min_Energy) Min_Energy = Energy;
  
         if ( Sweep > NTherm )
         {
            Total_Energy += Energy;
            Total_Magnet += Magnet;
            Total_Accept += Accept;

            Tot_Energy_Sq += Energy * Energy;
            Tot_Magnet_Sq += Magnet * Magnet;
            Tot_Accept_Sq += Accept * Accept;
         }
      }
 
      Measurements = Num_Sweeps - NTherm;
      Total_Energy /= Measurements;
      Total_Magnet /= Measurements;
      Total_Accept /= Measurements;

      Tot_Energy_Sq /= Measurements;
      Tot_Magnet_Sq /= Measurements;
      Tot_Accept_Sq /= Measurements;

      Error_Energy = Tot_Energy_Sq - (Total_Energy * Total_Energy);
      Error_Magnet = Tot_Magnet_Sq - (Total_Magnet * Total_Magnet);
      Error_Accept = Tot_Accept_Sq - (Total_Accept * Total_Accept);
   
      Error_Energy /= Num_Spins;
      Error_Magnet /= Num_Spins;
      Error_Accept /= Num_Spins;

      printf("%.8f\t%.8f\t%.8f\n",
             Beta_array[Cur_Bet_Ptr], Total_Energy, Error_Energy );
      fflush(stdout);

      Avge_Energy[Cur_Bet_Ptr] = Total_Energy;
      G_Diff += Total_Energy;
    } /* end of loop over betas */

    G_Diff = (G_Diff - 0.5*(Avge_Energy[0]+Avge_Energy[n])) * Beta_Diff;

    fprintf ( Outfile, "%.8f\t%.8f\n", Beta, G_Diff );
    fflush(Outfile);
    Beta_Diff = Beta_array[n] - Beta_array[0];
    Beta = Beta + Beta_Diff;
  }
  while (Beta < Beta_End);

  fprintf ( Outfile, "%.8f\t0.0\n", Beta );

  fprintf (Outfile, "\nMinimum Energy = %10.8lf\n", Min_Energy ); 

  fclose ( Outfile );

}       
           
/* PROCEDURE INITIALIZE_LATTICE */
void  Initialize_Lattice( Rows, Cols, Initial_Conf, Model )
int Rows, Cols, Initial_Conf, Model;
{
   int i, j, dim;

   switch (Initial_Conf)
   {

      case 1  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = 1;
	   printf("All spins 1\n");
           break;

      case 2  : 
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = -1;
	   printf("All spins -1\n");
           break;

      case 3  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 if ( Rand() < 0.5  )
                    Spins[i][j] = 1;
                 else 
                    Spins[i][j] = -1;
	   printf("Random spins\n");
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }

   switch (Model)
   {

      case 1  :  /* ferromagnet */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    J[i][j][dim] = 1;
	   printf("Ferromagnet\n");
           break;

      case 2  :  /* spin glass */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    if ( Rand() < 0.5  )
                       J[i][j][dim] = 1;
                    else 
                       J[i][j][dim] = -1;
	   printf("Spin Glass\n");
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }

}

/* PROCEDURE CALCULATE */
void  Calculate ( Rows, Cols, Energy, Magnet )
int Rows, Cols;
double *Energy, *Magnet;

{
   int Sum_Spins;
   int i, j;
   int ipos, jpos;

   *Magnet = 0.0;
   *Energy = 0.0;

   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
 
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
               
         Sum_Spins = Spins[i][j] * ( J[i][j][0] * Spins[ipos][j] + 
				     J[i][j][1] * Spins[i][jpos] );

         *Magnet += (double) Spins[i][j];
         *Energy += -1.0 * (double) Sum_Spins;
      }
   }

}

/* PROCEDURE METROPOLIS_UPDATE */
void  Metropolis_Update ( Beta, Rows, Cols, Accept )
double Beta;
int Rows, Cols;
double *Accept;
{
   int Old_Spin, New_Spin, Spin_Sum;
   double Old_Energy, New_Energy, Energy_Diff;
   int Acceptance;
   int i, j;
   int ipos, jpos, ineg, jneg;

   *Accept = 0.0;
   
   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
      if (i > 0)
         ineg = i-1;
      else 
         ineg = Rows-1;
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
         if (j > 0) 
            jneg = j-1;
         else 
            jneg = Cols-1;

         Acceptance = 0;

         Old_Spin = Spins[i][j];
         New_Spin = -Old_Spin;

         Spin_Sum = J[i][j][0]    * Spins[ipos][j] + 
		    J[i][j][1]    * Spins[i][jpos] + 
		    J[ineg][j][0] * Spins[ineg][j] + 
		    J[i][jneg][1] * Spins[i][jneg] ;

         Old_Energy = -Beta * Old_Spin * Spin_Sum;
         New_Energy = -Beta * New_Spin * Spin_Sum;
         Energy_Diff = New_Energy - Old_Energy;

         if ( (Energy_Diff <= 0.0) || ( exp(-Energy_Diff) > Rand() ) )
         { 
            Spins[i][j] = New_Spin;
            Acceptance = Acceptance + 1;
         }

         *Accept += (double)Acceptance;
      }
   }
}

/* PROCEDURE INPUT_PARAMETERS */
void Input_Parameters( Num_Spins, Rows, Cols, Initial_Conf, 
		       NTherm, Freq, Num_Sweeps, Beta_Start, Beta_End,
		       Beta_Factor, n, Seed, Model )
int *Num_Spins, *Rows, *Cols, *Initial_Conf;
int *NTherm, *Freq, *Num_Sweeps;
double *Beta_Start, *Beta_End, *Beta_Factor;
int *n, *Seed, *Model;
{
   int i; 
   char Sentinel;

   do   
   {
      Sentinel = 0;
      printf ( "Enter the vertical dimensions of the lattice (0 < Y <= %d): ",
	       MAXROWS );
      scanf ( "%d", Rows );

      if ( ( *Rows > MAXROWS ) || ( *Rows <= 0 ) )
      { 
         printf ( "\n\nInput out of range.\n\n" );
         Sentinel = 1;
      }
   } 
   while (Sentinel);
   printf("%d", *Rows);

   do
   {
      Sentinel = 0;

      printf("\n\n");
      printf ( "Enter the horizontal dimensions of the lattice (0 < X <= %d): ",
	       MAXCOLS );
      scanf ( "%d", Cols );

      if ( ( *Cols > MAXCOLS ) || ( *Cols <= 0 ) )
      {
            printf ( "\n\nInput out of range.\n\n" );
            Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Cols);

   *Num_Spins  = (*Rows) * (*Cols);
   printf ( "\n\nLATTICE SIZE :    %d X %d = %d\n", *Rows, *Cols, *Num_Spins );

   printf ( "\n\nEnter the number of thermalization sweeps desired: " );
   scanf ( "%d", NTherm );
   if ( *NTherm < 0 ) *NTherm = 0;
   printf("%d", *NTherm);

   do
   {
      Sentinel = 'n';        

      printf ( "\nEnter the total number of calculation updates \n" );
      printf ( "desired for each Beta: : " );
      scanf ( "%d", Num_Sweeps );
      if ( *Num_Sweeps <= 0 )   
         {
            printf ( "\n\nInvalid input\n\n" );
            Sentinel = 'y';
         }
   } while (Sentinel == 'y');
   printf("%d", *Num_Sweeps);

   do
   { 
      Sentinel = 'n';

      printf ( "\nEnter the number of sweeps per calculation update: ");
      scanf ( "%d", Freq );
      if ( *Freq > *Num_Sweeps ) *Freq = *Num_Sweeps;
      if ( *Freq <= 0 ) 
      { 
         printf ( "\n\nInvalid input\n\n" );
         Sentinel = 'y';
      }   
   } while ( Sentinel == 'y' ); 
   printf("%d", *Freq);

   do
   {  
      printf ( "\nEnter the Starting&Ending Beta value: " );
      scanf ( "%lf%lf", Beta_Start, Beta_End );
   }
   while ( *Beta_Start >= *Beta_End );
   printf("%f\t%f", *Beta_Start, *Beta_End);

    do
   {  
      printf ( "\nEnter the Beta Schedule Factor: " );
      scanf ( "%lf", Beta_Factor );
   }
   while ( *Beta_Factor <= 0 );
   printf("%f", *Beta_Factor);

   do
   {  
      printf ( "\nEnter the Grain Size for Calculating Delta_G: " );
      scanf ( "%d", n );
   }
   while ( *n <= 0 );
   printf("%d", *n);

   do 
   {
      Sentinel = 0;
      printf ( "\n\nEnter the number next to your choice of spin\n" );
      printf ( "arrangements for the initial lattice.\n\n" );
      printf ( "1 -- All spins up.\n" );
      printf ( "2 -- All spins down.\n" );
      printf ( "3 -- Random configuration.\n\n\n" );
      printf ( "Choice: " );
      scanf ( "%d", Initial_Conf );

      if ((*Initial_Conf != 1) && (*Initial_Conf != 2) && (*Initial_Conf != 3))
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Initial_Conf);
   
   do 
   {
      Sentinel = 0;
      printf ( "\n\nWhich model do you want to simulate?\n\n" );
      printf ( "1 -- Ferromagnet.\n" );
      printf ( "2 -- Spin Glass.\n" );
      printf ( "Choice: " );
      scanf ( "%d", Model );

      if ( (*Model != 1) && (*Model != 2) )
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Model);

   printf ( "\n\nInput a seed for the random number generator: " );
   scanf ( "%d", Seed );
   printf("%d\n\n", *Seed);
   fflush(stdout);

}



/* RANDOM NUMBER GENERATOR

 Standard Unix random number generator RAND.
 Generates random integers s_{i} in [0,MAXINT) using:
 s_{i+1} = (MULT * s_{i} + ADD) mod MAXINT
*/

#define	MULT	1103515245
#define	ADD 	12345
#define	MASK	( 0x7fffffff )
#define	MAXINT	( (double)MASK + 1.0 )      /* 2^{31} */
long randx;     /* s_{i} */

/*
	Initialize the rng.
*/
Rand_Init(seed)
int seed;
{
	randx = (MULT*seed + ADD) & MASK;
}


/*
	Return a random float in [0, 1.0)
*/
float Rand()
{
	float retvalue;

	retvalue = randx / MAXINT;
	randx = (MULT*randx + ADD) & MASK;
	if (retvalue == 1.0) retvalue = 0.0;
	return( retvalue );
}

