/***************************************************/
/* Student : Chien-Hsing Wu                        */
/* S. S. No : 105-76-3704                          */
/* File name : temper.c                            */
/* Compile Instruction : cc temper.c -o temper -lm */
/* Run instruction : temper Infofile	           */
/* This program accepts a list of Beta's and Delta */
/* gm's and collect the acceptance rates of each   */
/* Beta's.					   */
/***************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define MAXROWS 256
#define MAXCOLS 256
#define MAXBETAS 50

#define UP 0
#define DOWN 1
#define DIRS 2

#define TRUE 1
#define FALSE 0

int Spins[MAXROWS][MAXCOLS];    /* spins */
int J[MAXROWS][MAXCOLS][2];     /* couplings */
int Beta_sweeps[MAXBETAS];

double Beta_array[MAXBETAS];
double Delta_G[MAXBETAS];

int Rows, Cols, Num_Spins;

float Rand();


main(argc, argv)
int  argc;
char  *argv[];
{
   int Sweep;
   int Lattice;  
   int NTherm;
   int Cur_Bet_Ptr;
   int Num_Betas;
   int Initial_Conf;     
   int Freq;          
   int Num_Sweeps;     
   int Model;
   int Seed;
   int Beta_Accept[MAXBETAS][DIRS];
   int Beta_Calcs[MAXBETAS][DIRS];
     
   double Beta;          
   double Energy;       
   double Magnet;         
   double Accept; 
   double Min_Energy = 100.0;   

   void Input_Parameters ();
   void Metropolis_Update ();
   void Beta_Update ();
   void Initialize_Lattice (); 
   void Calculate ();

   FILE *Outfile;

  if (argc != 2)
  {
    printf("Usage : temper Infofile\n");
    exit(1);
  }

  if ((Outfile = fopen(argv[1], "w")) == NULL)
  {
    printf("\nError Opening the Infofile : %s\n", argv[1]);
    exit(2);
  }
   
   Input_Parameters( &Num_Spins, &Rows, &Cols, &Initial_Conf,
		     &NTherm, &Freq, &Num_Sweeps, &Num_Betas, &Seed, &Model );

   /* Initialize random numbers */
   Rand_Init( Seed );

   Initialize_Lattice ( Rows, Cols, Initial_Conf, Model );

   Calculate ( Rows, Cols, &Energy, &Magnet );

   for ( Sweep = 0; Sweep < NTherm; Sweep++ )
   {
         Beta = Beta_array[0];
         Metropolis_Update ( Beta, Rows, Cols, &Accept );
   }

   Calculate ( Rows, Cols, &Energy, &Magnet );
   
   for ( Cur_Bet_Ptr = 0; Cur_Bet_Ptr < Num_Betas; Cur_Bet_Ptr++ )
   {
      Beta_Accept[Cur_Bet_Ptr][UP] = 0.0;
      Beta_Accept[Cur_Bet_Ptr][DOWN] = 0.0;
      Beta_Calcs[Cur_Bet_Ptr][UP] = 0.0;
      Beta_Calcs[Cur_Bet_Ptr][DOWN] = 0.0;
   }

   Cur_Bet_Ptr = 0;
   for ( Sweep = 0; Sweep < Num_Sweeps; Sweep++ )
   {

      Beta_Update ( &Cur_Bet_Ptr, Num_Betas, Energy, Beta_Accept, Beta_Calcs );
      Beta_sweeps[Cur_Bet_Ptr] += 1;
      Beta = Beta_array[Cur_Bet_Ptr];

      for ( Lattice = 0; Lattice < Freq; Lattice++ )
      {
         Metropolis_Update ( Beta, Rows, Cols, &Accept );
      }
 
      Calculate ( Rows, Cols, &Energy, &Magnet );

      if (Energy < Min_Energy) Min_Energy = Energy;

      fprintf ( Outfile, "%6d  %10.6lf  %f\n", Sweep+1,  Energy, Beta ); 
      fflush ( Outfile );

   }

   for ( Cur_Bet_Ptr = 0; Cur_Bet_Ptr < Num_Betas-1; Cur_Bet_Ptr++ )
   {
       if (Beta_Calcs[Cur_Bet_Ptr+1][UP] == 0)
          Beta_Calcs[Cur_Bet_Ptr+1][UP] = 1.0;
       if (Beta_Calcs[Cur_Bet_Ptr][DOWN] == 0)
          Beta_Calcs[Cur_Bet_Ptr][DOWN] = 1.0;

       fprintf ( Outfile,   "%f\t%10.8lf\t%10.8lf\n", 
		            Beta_array[Cur_Bet_Ptr],
			    (double)Beta_Accept[Cur_Bet_Ptr+1][UP] / 
			    (double)Beta_Calcs[Cur_Bet_Ptr+1][UP], 
			    (double)Beta_Accept[Cur_Bet_Ptr][DOWN] / 
			    (double)Beta_Calcs[Cur_Bet_Ptr][DOWN]);
   }
   
   for ( Cur_Bet_Ptr = 0; Cur_Bet_Ptr < Num_Betas; Cur_Bet_Ptr++ )
   {
	fprintf( Outfile, "%f\t%d\n", Beta_array[Cur_Bet_Ptr],
				      Beta_sweeps[Cur_Bet_Ptr]);
   }
   printf ("\n\nMinimum Energy = %10.8lf\n", Min_Energy ); 

   fclose ( Outfile );

}       
           
/* PROCEDURE INITIALIZE_LATTICE */
void  Initialize_Lattice( Rows, Cols, Initial_Conf, Model )
int Rows, Cols, Initial_Conf, Model;
{
   int i, j, dim;

   switch (Initial_Conf)
   {

      case 1  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = 1;
	   printf("All spins 1\n");
           break;

      case 2  : 
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 Spins[i][j] = -1;
	   printf("All spins -1\n");
           break;

      case 3  :
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 if ( Rand() < 0.5  )
                    Spins[i][j] = 1;
                 else 
                    Spins[i][j] = -1;
	   printf("Random spins\n");
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }

   switch (Model)
   {

      case 1  :  /* ferromagnet */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    J[i][j][dim] = 1;
	   printf("Ferromagnet\n");
           break;

      case 2  :  /* spin glass */
           for ( i = 0; i < Rows; i++ )
              for ( j = 0; j < Cols; j++ )           
                 for ( dim = 0; dim < 2; dim++ )           
                    if ( Rand() < 0.5  )
                       J[i][j][dim] = 1;
                    else 
                       J[i][j][dim] = -1;
	   printf("Spin Glass\n");
           break;
 
      default  :
           printf("initialization of spins is screwed up!\n");
	   fflush(stdout);

   }

}

/* PROCEDURE CALCULATE */
void
Calculate ( Rows, Cols, Energy, Magnet )
int Rows, Cols;
double *Energy, *Magnet;

{
   int Sum_Spins;
   int i, j;
   int ipos, jpos;

   *Magnet = 0.0;
   *Energy = 0.0;

   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
 
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
               
         Sum_Spins = Spins[i][j] * ( J[i][j][0] * Spins[ipos][j] + 
				     J[i][j][1] * Spins[i][jpos] );

         *Magnet += (double) Spins[i][j];
         *Energy += -1.0 * (double) Sum_Spins;
      }
   }

}

/* PROCEDURE METROPOLIS_UPDATE */
void  Metropolis_Update ( Beta, Rows, Cols, Accept )
double Beta;
int Rows, Cols;
double *Accept;
{
   int Old_Spin, New_Spin, Spin_Sum;
   double Old_Energy, New_Energy, Energy_Diff;
   int Acceptance;
   int i, j;
   int ipos, jpos, ineg, jneg;

   *Accept = 0.0;
   
   for ( i = 0; i < Rows; i++ )
   {
      if (i < Rows-1) 
         ipos = i+1;
      else 
         ipos = 0;
      if (i > 0)
         ineg = i-1;
      else 
         ineg = Rows-1;
      for ( j = 0; j < Cols; j++ ) 
      {
         if (j < Cols-1)
            jpos = j+1;
         else 
            jpos = 0;
         if (j > 0) 
            jneg = j-1;
         else 
            jneg = Cols-1;

         Acceptance = 0;

         Old_Spin = Spins[i][j];
         New_Spin = -Old_Spin;

         Spin_Sum = J[i][j][0]    * Spins[ipos][j] + 
		    J[i][j][1]    * Spins[i][jpos] + 
		    J[ineg][j][0] * Spins[ineg][j] + 
		    J[i][jneg][1] * Spins[i][jneg] ;

         Old_Energy = -Beta * Old_Spin * Spin_Sum;
         New_Energy = -Beta * New_Spin * Spin_Sum;
         Energy_Diff = New_Energy - Old_Energy;

         if ( (Energy_Diff <= 0.0) || ( exp(-Energy_Diff) > Rand() ) )
         { 
            Spins[i][j] = New_Spin;
            Acceptance = Acceptance + 1;
         }

         *Accept += (double)Acceptance;
      }
   }
}

/* PROCEDURE BETA_UPDATE */
void
Beta_Update ( Cur_Bet_Ptr, Num_Betas, Energy, Accept_Beta, Calcs_Beta )
int *Cur_Bet_Ptr;
int Num_Betas;
double Energy;
int Accept_Beta[MAXBETAS][DIRS], Calcs_Beta[MAXBETAS][DIRS];
{
   int New_Bet_Ptr, Old_Bet_Ptr, Dir, Update;
   double Boltz_Diff, Beta_Diff, G_Diff;

   Old_Bet_Ptr = *Cur_Bet_Ptr;
   /* Choose new neighboring beta value */
   if ( Rand() < 0.5 ) {
      New_Bet_Ptr = Old_Bet_Ptr + 1;
      Dir = DOWN;
      if ( New_Bet_Ptr >= Num_Betas) {
         New_Bet_Ptr = Num_Betas - 1;
         Update = FALSE;
      }
      else {
         Update = TRUE;
         Calcs_Beta[Old_Bet_Ptr][Dir] += 1;
      }    
      G_Diff = Delta_G[Old_Bet_Ptr];
   }
   else {
      New_Bet_Ptr = Old_Bet_Ptr - 1;
      Dir = UP;
      if ( New_Bet_Ptr < 0 ) {
         New_Bet_Ptr = 0;
         Update = FALSE;
      }
      else {
         Update = TRUE;
         Calcs_Beta[Old_Bet_Ptr][Dir] += 1;
      }
      G_Diff = - Delta_G[New_Bet_Ptr];
   }

   Beta_Diff = Beta_array[New_Bet_Ptr] - Beta_array[Old_Bet_Ptr];
   Boltz_Diff = (Energy * Beta_Diff) - G_Diff;

   if ( ( Boltz_Diff <= 0.0 ) || ( exp ( -Boltz_Diff) > Rand() ) ) {
            *Cur_Bet_Ptr = New_Bet_Ptr;
	    Accept_Beta[Old_Bet_Ptr][Dir] += Update;
   }
}

/* PROCEDURE INPUT_PARAMETERS */
void  Input_Parameters( Num_Spins, Rows, Cols, Initial_Conf, 
		  NTherm, Freq, Num_Sweeps, Num_Betas, Seed, Model )
int *Num_Spins, *Rows, *Cols, *Initial_Conf;
int *NTherm, *Freq, *Num_Sweeps;
int *Num_Betas;
int *Seed, *Model;

{
   int i; 
   char Sentinel;

   do   
   {
      Sentinel = 0;
      printf ( "Enter the vertical dimensions of the lattice (0 < Y <= %d): ",
	       MAXROWS );
      scanf ( "%d", Rows );

      if ( ( *Rows > MAXROWS ) || ( *Rows <= 0 ) )
      { 
         printf ( "\n\nInput out of range.\n\n" );
         Sentinel = 1;
      }
   } 
   while (Sentinel);
   printf("%d", *Rows);

   do
   {
      Sentinel = 0;

      printf("\n\n");
      printf ( "Enter the horizontal dimensions of the lattice (0 < X <= %d): ",
	       MAXCOLS );
      scanf ( "%d", Cols );

      if ( ( *Cols > MAXCOLS ) || ( *Cols <= 0 ) )
      {
            printf ( "\n\nInput out of range.\n\n" );
            Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Cols);

   *Num_Spins  = (*Rows) * (*Cols);
   printf ( "\n\nLATTICE SIZE :    %d X %d = %d\n", *Rows, *Cols, *Num_Spins );

   printf ( "\n\nEnter the number of thermalization sweeps desired: " );
   scanf ( "%d", NTherm );
   if ( *NTherm < 0 ) *NTherm = 0;
   printf("%d", *NTherm);

   do
   {
      Sentinel = 'n';        

      printf ( "\nEnter the total number of calculation updates \n" );
      printf ( "desired for each Beta: : " );
      scanf ( "%d", Num_Sweeps );
      if ( *Num_Sweeps <= 0 )   
         {
            printf ( "\n\nInvalid input\n\n" );
            Sentinel = 'y';
         }
   } while (Sentinel == 'y');
   printf("%d", *Num_Sweeps);

   do
   { 
      Sentinel = 'n';

      printf ( "\nEnter the number of sweeps per calculation update: ");
      scanf ( "%d", Freq );
      if ( *Freq > *Num_Sweeps ) *Freq = *Num_Sweeps;
      if ( *Freq <= 0 ) 
      { 
         printf ( "\n\nInvalid input\n\n" );
         Sentinel = 'y';
      }   
   } while ( Sentinel == 'y' ); 
   printf("%d", *Freq);

    printf ( "\nEnter the Beta values in order.\n" );
    printf ( "Input -1 to end input. ( # values <= %d )\n\n", MAXBETAS );

    *Num_Betas = 0;
    i = 0;
    do
    {
       scanf ( "%lf", Beta_array + i );
       (*Num_Betas)++;
       i++;
    } 
    while ( *( Beta_array + i - 1 ) > 0.0 );

    (*Num_Betas)--;
     
    printf ( "CURRENT BETA VALUES\n" );
    printf ( "-------------------\n" );
 
    for ( i = 0; i < *Num_Betas; i++ ) 
    {
       printf ( "%14.8lf\n", *( Beta_array + i ) );
    }
    fflush(stdout);
     
    printf ( "\n\nCURRENT G_DIFF VALUES\n" );
    printf ( "-------------  ------\n" );
 
    for ( i = 0; i < *Num_Betas; i++ ) 
    {
       scanf ( "%lf", Delta_G + i );
       printf ( "%14.8lf\n", *( Delta_G + i ) );
    }
 
   do 
   {
      Sentinel = 0;
      printf ( "\n\nEnter the number next to your choice of spin\n" );
      printf ( "arrangements for the initial lattice.\n\n" );
      printf ( "1 -- All spins up.\n" );
      printf ( "2 -- All spins down.\n" );
      printf ( "3 -- Random configuration.\n\n\n" );
      printf ( "Choice: " );
      scanf ( "%d", Initial_Conf );

      if ((*Initial_Conf != 1) && (*Initial_Conf != 2) && (*Initial_Conf != 3))
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Initial_Conf);
   
   do 
   {
      Sentinel = 0;
      printf ( "\n\nWhich model do you want to simulate?\n\n" );
      printf ( "1 -- Ferromagnet.\n" );
      printf ( "2 -- Spin Glass.\n" );
      printf ( "Choice: " );
      scanf ( "%d", Model );

      if ( (*Model != 1) && (*Model != 2) )
      {
         printf ( "\n\nInvalid input\n" );
         Sentinel = 1;
      }
   }
   while (Sentinel);
   printf("%d", *Model);

   printf ( "\n\nInput a seed for the random number generator: " );
   scanf ( "%d", Seed );
   printf("%d\n\n", *Seed);
   fflush(stdout);

}




/* RANDOM NUMBER GENERATOR

 Standard Unix random number generator RAND.
 Generates random integers s_{i} in [0,MAXINT) using:
 s_{i+1} = (MULT * s_{i} + ADD) mod MAXINT
*/

#define	MULT	1103515245
#define	ADD 	12345
#define	MASK	( 0x7fffffff )
#define	MAXINT	( (double)MASK + 1.0 )      /* 2^{31} */
long randx;     /* s_{i} */

/*
	Initialize the rng.
*/
Rand_Init(seed)
int seed;
{
	randx = (MULT*seed + ADD) & MASK;
}


/*
	Return a random float in [0, 1.0)
*/
float Rand()
{
	float retvalue;

	retvalue = randx / MAXINT;
	randx = (MULT*randx + ADD) & MASK;
	if (retvalue == 1.0) retvalue = 0.0;
	return( retvalue );
}

