
/***********************************************************************
 *                                                                     *
 *          This software copyrighted by Caltech, 1988.                *
 *                                                                     *
 *          Author:  Ed Felten                                         *
 *                                                                     *
 ***********************************************************************/




#include <stdio.h>
#include "opt.h"


int procnum,nprocs;

int graphics = FALSE;


main()
{
    char    str[80];
    int     npts,i;
    int     done;
    int     stime;
    int     tour[MAXCITIES];
    int     dist[MAXCITIES][MAXCITIES];
    int     coord[MAXCITIES][2];
    float   start_temp,end_temp;

    procnum = 0;
    nprocs = 1;
    ranset(173);
    init_my_exp();
    
    npts = get_data(dist,coord);

    do{
        done = FALSE;
	printf("What now?\n");
	scanf("%s",str);
	switch(str[0]){
            case 'a' :
		printf("How many hits?\n");
		scanf("%d",&i);
		printf("Enter starting and ending temperatures:\n");
		scanf("%f %f",&start_temp,&end_temp);
                monte_carlo(npts,tour,dist,coord,i,start_temp,end_temp);
		continue;
	    case 'g' :
		if(graphics){
		    graphics = FALSE;
		    printf("Turning graphics OFF.\n");
		}
		else{
		    graphics = TRUE;
		    printf("Turning graphics On.\n");
		}
		continue;
	    case 'i' :
		random_tour(npts,tour);
		continue;
	    case 'l' :
		printf("Tour length is %d\n",tour_length(npts,tour,dist));
		continue;
            case 'r' :
                printf("Enter new random number seed:\n");
                scanf("%d",&i);
                ranset(i);
                continue;
            case 'R' :
                printf("What file should I read the tour from?");
                scanf("%s",str);
                read_tour(npts,tour,str);
                continue;
	    case 's' :
		printf("How many links should I scramble?\n");
		scanf("%d",&i);
		scramble(npts,tour,i);
		continue;
	    case 't' :
		if(graphics){
		    plot(npts,tour,coord);
		}
		else{
		    printf("Sorry, graphics are turned off.\n");
		}
		continue;
	    case 'v' :
		if( verify_tour(npts,tour) ){
		    printf("Tour is otay!\n");
		}
		else{
		    printf("Tour has been corrupted!\n");
		}
		continue;
            case 'W' :
                printf("What file should I write the tour to?");
                scanf("%s",str);
                write_tour(npts,tour,str);
                continue;
            case 'x' :
                done = TRUE;
                break;
	    case '3' :
		three_opt(npts,tour,dist,coord,PLOT_EVERY);
		continue;
	    default:
		help();
		continue;
	}
    }while(! done);
        
    exit(0);
}


help()
{
    printf("\nHere's what you can do:\n");
    printf("a		anneal using super monte carlo\n");
    printf("g		turn graphics on or off\n");
    printf("i		initialize tour randomly\n");
    printf("l		measure length of tour\n");
    printf("r		reseed random number generator\n");
    printf("R		Read tour from disk\n");
    printf("s		scramble the tour by a random k-change\n");
    printf("t		print the tour\n");
    printf("v		verify that the tour is feasible\n");
    printf("W		Write tour to disk\n");
    printf("x		exit from simulation\n");
    printf("3		3-opt the tour\n");
    printf("?		examine this list\n");
}


int tour_length(npts,tour,dist)
    int     npts;
    int     tour[];
    int     dist[MAXCITIES][MAXCITIES];
{
    int     i;
    int     ret;

    ret = dist[tour[npts-1]][tour[0]];
    for(i=0; i<npts-1; ++i){
	ret += dist[tour[i]][tour[i+1]];
    }

    return(ret);
}



int get_data(dist,coord)
    int     dist[MAXCITIES][MAXCITIES];
    int     coord[MAXCITIES][2];
{
    char    name[80];
    int     npts,i,j;
    double  dx,dy;
    FILE    *fp;

    FILE    *fopen();
    double  sqrt();

    printf("How many cities?\n");
    scanf("%d",&npts);
    printf("Problem has %d cities.\n",npts);

    printf("Filename for coordinates?\n");
    scanf("%s",name);

    fp = fopen(name,"r");
    if(fp == NULL){
        fprintf("Couldn't open %s for reading.\n",name);
        exit(1);
    }

    for(i=0; i<npts; ++i){
	fscanf(fp,"%d %d",&coord[i][0],&coord[i][1]);
    }
    fclose(fp);

    /*  calculate distance matrix  */
    for(i=0; i<npts; ++i){
        for(j=i; j<npts; ++j){
            dx = coord[i][0]-coord[j][0];
            dy = coord[i][1]-coord[j][1];
            dist[j][i] = dist[i][j] = (int)(sqrt(dx*dx+dy*dy));
        }
    }

    return(npts);
}


random_tour(npts,tour)
    int     npts;
    int     tour[];
{
    int     i,temp,other_guy;

    for(i=0; i<npts; ++i){
	tour[i] = i;
    }
    for(i=npts-1; i>0; --i){
	other_guy = random_int(i+1);
	temp = tour[other_guy];
	tour[other_guy] = tour[i];
	tour[i] = temp;
    }
}


int random_int(max)
    int     max;
{
    return( (rand()>>10) % max );
}


print_tour(npts,tour)
    int     npts,tour[];
{
    int     i;

    for(i=0; i<npts; ++i){
        printf("%d ",tour[i]);
    }
    printf("\n");
}


print_time(stime)
    int     stime;
{
    printf("%.2f seconds",stime*1024.0/6.9e6);
}

int verify_tour(npts,tour)
    int     npts;
    int     tour[];
{
    int     i,agree;
    int     proc0_tour[MAXCITIES];
    int     degree[MAXCITIES];

    int     and_int();

    if(procnum == 0){
/*
	broadcast(tour,0,nprocs-1,npts*sizeof(int));
*/
	agree = TRUE;
    }
    else{
/*
	broadcast(proc0_tour,0,nprocs-1,npts*sizeof(int));
*/
	agree = TRUE;
	for(i=0; i<npts; ++i){
	    if(proc0_tour[i] != tour[i]){
		agree = FALSE;
	    }
	}
    }
/*
    combine(&agree,and_int,sizeof(int),1);
*/
    if( ! agree ){
	printf("Processors have different tours!\n");
	return(FALSE);
    }

    
    for(i=0; i<npts; ++i){
        degree[i] = 0;
    }

    for(i=0; i<npts; ++i){
	if( (tour[i]<0) || (tour[i]>=npts) ){
	    return(FALSE);
	}
	++degree[tour[i]];
	if(degree[tour[i]] > 1){
	    return(FALSE);
	}
    }

    /*  nothing went wrong, tour must be otay  */
    return(TRUE);
}


int and_int(ip1,ip2,size)
    int     *ip1,*ip2;
    int     size;
{
    *ip1 &= *ip2;
}


read_tour(npts,tour,filename)
    int     npts;
    int     tour[];
    char    filename[];
{
    int     i;
    FILE    *fp;

    FILE    *fopen();

    fp = fopen(filename,"r");
    if(fp == NULL){
	fprintf(stderr,"Couldn't open %s for reading.\n",filename);
	fflush(stderr);
	return;
    }

    for(i=0; i<npts; ++i){
	fscanf(fp,"%d",&tour[i]);
    }
    fclose(fp);
}


write_tour(npts,tour,filename)
    int     npts;
    int     tour[];
    char    filename[];
{
    int     i;
    FILE    *fp;

    FILE    *fopen();

    fp = fopen(filename,"w");
    if(fp == NULL){
	fprintf(stderr,"Couldn't open %s for writing.\n",filename);
	fflush(stderr);
	return;
    }

    for(i=0; i<npts; ++i){
	fprintf(fp,"%d\n",tour[i]);
    }
    fclose(fp);
}
