

/***********************************************************************
 *                                                                     *
 *          This software copyrighted by Caltech, 1988.                *
 *                                                                     *
 *          Author:  Ed Felten                                         *
 *                                                                     *
 ***********************************************************************/



#include <stdio.h>
#include "opt.h"


extern int procnum,nprocs;

extern int graphics;

typedef struct {
    int     i,j,k;
    int     rev_flag;
    int     improvement;
} THREE_CHANGE;

#define DIST(I,J)    dist[tour[I]][tour[J]]


void three_opt(npts,tour,dist,coord,plot_when)
    int     npts;
    int     tour[];
    int     dist[MAXCITIES][MAXCITIES];
    int     coord[MAXCITIES][2];
    int     plot_when;
{
    int     i,j,k;
    int     loops_each;
    int     old,forward,backward,improvement;
    THREE_CHANGE try;

    int     pick_best();

    loops_each = (npts+nprocs-1)/nprocs;

    do{
        if(plot_when == PLOT_EVERY){
            printf("%d\n",tour_length(npts,tour,dist));
	    if(graphics){
                plot(npts,tour,coord);
	    }
        }
        try.improvement = -1000000;
        for(i=procnum*loops_each; (i<(procnum+1)*loops_each)&&(i<npts); ++i){
	    for(j=(i+1)%npts; j!=i; j=(j+1)%npts){
	        for(k=(j+1)%npts; k!=i; k=(k+1)%npts){
		    old = DIST(i,(i+1)%npts)+DIST(j,(j+1)%npts)+
                                             DIST(k,(k+1)%npts);
		    forward = DIST(i,(j+1)%npts)+DIST(k,(i+1)%npts)+
                                                 DIST(j,(k+1)%npts);
		    backward = DIST(i,(j+1)%npts);
                    backward += DIST(k,j);
		    backward += DIST((i+1)%npts,(k+1)%npts);
		    if(forward <= backward){
		        improvement = old-forward;
		        if(improvement > try.improvement){
			    try.improvement = improvement;
			    try.i = i;
			    try.j = j;
			    try.k = k;
			    try.rev_flag = FALSE;
		        }
		    }
		    else{
		        improvement = old-backward;
		        if(improvement > try.improvement){
			    try.improvement = improvement;
			    try.i = i;
			    try.j = j;
			    try.k = k;
			    try.rev_flag = TRUE;
		        }
		    }
	        }
	    }
        }

/*
        combine(&try,pick_best,sizeof(THREE_CHANGE),1);
*/    
	if(try.improvement > 0){
	    three_change(npts,tour,try.i,try.j,try.k,try.rev_flag);
	}
    }while( try.improvement > 0 );

    if( (plot_when==PLOT_END) && (plot_when!=PLOT_EVERY) ){
        printf("%d\n",tour_length(npts,tour,dist));
	if(graphics){
            plot(npts,tour,coord);
	}
    }
}



int pick_best(pt1,pt2,size)
    THREE_CHANGE *pt1,*pt2;
    int     size;
{
    if(pt2->improvement > pt1->improvement){
        *pt1 = *pt2;
    }else if(pt2->improvement == pt1->improvement){
	if(pt2->i < pt1->i){
	    *pt1 = *pt2;
	}
    } 

    return(0);
}
    

three_change(npts,tour,i,j,k,rev_flag)
    int     npts,tour[];
    int     i,j,k,rev_flag;
{
    int     new,old;
    int     temp_tour[MAXCITIES];

    for(new=0; new<npts; ++new){
	temp_tour[new] = tour[new];
    }

    new = 0;
    for(old=(i+1)%npts; old!=j; old=(old+1)%npts){
	tour[new++] = temp_tour[old];
    }
    tour[new++] = temp_tour[j];
    if(rev_flag){
	for(old=k; old!=(j+1)%npts; old=(old+npts-1)%npts){
	    tour[new++] = temp_tour[old];
	}
        tour[new++] = temp_tour[(j+1)%npts];
	for(old=i; old!=(k+1)%npts; old=(old+npts-1)%npts){
	    tour[new++] = temp_tour[old];
	}
        tour[new++] = temp_tour[(k+1)%npts];
    }
    else{
	for(old=(k+1)%npts; old!=i; old=(old+1)%npts){
	    tour[new++] = temp_tour[old];
	}
        tour[new++] = temp_tour[i];
	for(old=(j+1)%npts; old!=k; old=(old+1)%npts){
	    tour[new++] = temp_tour[old];
	}
        tour[new++] = temp_tour[k];
    }

    if(new != npts){
        fprintf(stderr,"Whoa!  new is %d\n",new);
        fflush(stderr);
        exit(1);
    }
}



